# This file is part of nvitop, the interactive NVIDIA-GPU process viewer.
# License: GNU GPL version 3.

# pylint: disable=missing-module-docstring

from nvitop.gui.library.device import NA, Device
from nvitop.gui.library.displayable import Displayable, DisplayableContainer
from nvitop.gui.library.history import BufferedHistoryGraph, HistoryGraph
from nvitop.gui.library.keybinding import (
    ALT_KEY,
    ANYKEY,
    PASSIVE_ACTION,
    QUANT_KEY,
    SPECIAL_KEYS,
    KeyBuffer,
    KeyMaps,
    normalize_keybinding,
)
from nvitop.gui.library.libcurses import libcurses, setlocale_utf8
from nvitop.gui.library.messagebox import MessageBox, send_signal
from nvitop.gui.library.mouse import MouseEvent
from nvitop.gui.library.process import (
    GiB,
    GpuProcess,
    HostProcess,
    Snapshot,
    bytes2human,
    host,
    timedelta2human,
)
from nvitop.gui.library.selection import Selection
from nvitop.gui.library.utils import (
    HOSTNAME,
    LARGE_INTEGER,
    SUPERUSER,
    USERCONTEXT,
    USERNAME,
    colored,
    cut_string,
    make_bar,
    set_color,
)
from nvitop.gui.library.widestring import WideString, wcslen



#!/usr/bin/env python3

# To install `nvitop` with specific version of `nvidia-ml-py`, use:
#
#   pip install nvidia-ml-py==xx.yyy.zz nvitop
#
# or
#
#   pip install 'nvitop[pynvml-xx.yyy.zz]'
#

"""Setup script for ``nvitop``."""

import pathlib
import re
import sys

from setuptools import setup


HERE = pathlib.Path(__file__).absolute().parent
VERSION_FILE = HERE / 'nvitop' / 'version.py'

sys.path.insert(0, str(VERSION_FILE.parent))
# pylint: disable-next=import-error,wrong-import-position
import version  # noqa


VERSION_CONTENT = None

try:
    if not version.__release__:
        try:
            VERSION_CONTENT = VERSION_FILE.read_text(encoding='utf-8')
            VERSION_FILE.write_text(
                data=re.sub(
                    r"""__version__\s*=\s*('[^']+'|"[^"]+")""",
                    f'__version__ = {version.__version__!r}',
                    string=VERSION_CONTENT,
                ),
                encoding='utf-8',
            )
        except OSError:
            VERSION_CONTENT = None

    setup(
        name='nvitop',
        version=version.__version__,
        extras_require={
            'lint': [
                'black >= 22.6.0',
                'isort',
                'pylint[spelling] >= 2.16.0',
                'mypy',
                'typing-extensions',
                'pre-commit',
            ],
            'cuda10': ['nvidia-ml-py == 11.450.51'],
            **{
                # The identifier could not start with numbers, add a prefix `pynvml-`
                f'pynvml-{pynvml}': [f'nvidia-ml-py == {pynvml}']
                for pynvml in version.PYNVML_VERSION_CANDIDATES
            },
        },
    )
finally:
    if VERSION_CONTENT is not None:
        with VERSION_FILE.open(mode='wt', encoding='utf-8', newline='') as file:
            file.write(VERSION_CONTENT)



# This file is part of nvitop, the interactive NVIDIA-GPU process viewer.
#
# Copyright 2021-2023 Xuehai Pan. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Python bindings for the `CUDA Driver APIs <https://docs.nvidia.com/cuda/cuda-driver-api>`_."""

# pylint: disable=invalid-name

from __future__ import annotations

import ctypes as _ctypes
import itertools as _itertools
import platform as _platform
import string as _string
import sys as _sys
import threading as _threading
from typing import TYPE_CHECKING as _TYPE_CHECKING
from typing import Any as _Any
from typing import Callable as _Callable
from typing import ClassVar as _ClassVar


if _TYPE_CHECKING:
    from typing_extensions import Self as _Self  # Python 3.11+
    from typing_extensions import TypeAlias as _TypeAlias  # Python 3.10+


# pylint: disable-next=missing-class-docstring,too-few-public-methods
class _struct_c_CUdevice_t(_ctypes.Structure):
    pass  # opaque handle


_c_CUdevice_t: _TypeAlias = _ctypes.POINTER(  # type: ignore[valid-type] # noqa: PYI042
    _struct_c_CUdevice_t,
)

_CUresult_t: _TypeAlias = _ctypes.c_uint

# Error codes #
# pylint: disable=line-too-long
CUDA_SUCCESS = 0
"""The API call returned with no errors. In the case of query calls, this also means that the operation being queried is complete (see :func:`cuEventQuery` and :func:`cuStreamQuery`)."""
CUDA_ERROR_INVALID_VALUE = 1
"""This indicates that one or more of the parameters passed to the API call is not within an acceptable range of values."""
CUDA_ERROR_OUT_OF_MEMORY = 2
"""The API call failed because it was unable to allocate enough memory to perform the requested operation."""
CUDA_ERROR_NOT_INITIALIZED = 3
"""This indicates that the CUDA driver has not been initialized with :func:`cuInit` or that initialization has failed."""
CUDA_ERROR_DEINITIALIZED = 4
"""This indicates that the CUDA driver is in the process of shutting down."""
CUDA_ERROR_PROFILER_DISABLED = 5
"""This indicates profiler is not initialized for this run. This can happen when the application is running with external profiling tools like visual profiler."""
CUDA_ERROR_STUB_LIBRARY = 34
"""This indicates that the CUDA driver that the application has loaded is a stub library. Applications that run with the stub rather than a real driver loaded will result in CUDA API returning this error."""
CUDA_ERROR_DEVICE_UNAVAILABLE = 46
"""This indicates that requested CUDA device is unavailable at the current time. Devices are often unavailable due to use of :data:`CU_COMPUTEMODE_EXCLUSIVE_PROCESS` or :data:`CU_COMPUTEMODE_PROHIBITED`."""
CUDA_ERROR_NO_DEVICE = 100
"""This indicates that no CUDA - capable devices were detected by the installed CUDA driver."""
CUDA_ERROR_INVALID_DEVICE = 101
"""This indicates that the device ordinal supplied by the user does not correspond to a valid CUDA device or that the action requested is invalid for the specified device."""
CUDA_ERROR_DEVICE_NOT_LICENSED = 102
"""This error indicates that the Grid license is not applied."""
CUDA_ERROR_INVALID_IMAGE = 200
"""This indicates that the device kernel image is invalid. This can also indicate an invalid CUDA module."""
CUDA_ERROR_INVALID_CONTEXT = 201
"""This most frequently indicates that there is no context bound to the current thread. This can also be returned if the context passed to an API call is not a valid handle (such as a context that has had :func:`cuCtxDestroy` invoked on it). This can also be returned if a user mixes different API versions (i.e. 3010 context with 3020 API calls). See :func:`cuCtxGetApiVersion` for more details."""
CUDA_ERROR_MAP_FAILED = 205
"""This indicates that a map or register operation has failed."""
CUDA_ERROR_UNMAP_FAILED = 206
"""This indicates that an unmap or unregister operation has failed."""
CUDA_ERROR_ARRAY_IS_MAPPED = 207
"""This indicates that the specified array is currently mapped and thus cannot be destroyed."""
CUDA_ERROR_ALREADY_MAPPED = 208
"""This indicates that the resource is already mapped."""
CUDA_ERROR_NO_BINARY_FOR_GPU = 209
"""This indicates that there is no kernel image available that is suitable for the device. This can occur when a user specifies code generation options for a particular CUDA source file that do not include the corresponding device configuration."""
CUDA_ERROR_ALREADY_ACQUIRED = 210
"""This indicates that a resource has already been acquired."""
CUDA_ERROR_NOT_MAPPED = 211
"""This indicates that a resource is not mapped."""
CUDA_ERROR_NOT_MAPPED_AS_ARRAY = 212
"""This indicates that a mapped resource is not available for access as an array."""
CUDA_ERROR_NOT_MAPPED_AS_POINTER = 213
"""This indicates that a mapped resource is not available for access as a pointer."""
CUDA_ERROR_ECC_UNCORRECTABLE = 214
"""This indicates that an uncorrectable ECC error was detected during execution."""
CUDA_ERROR_UNSUPPORTED_LIMIT = 215
"""This indicates that the :class:`CUlimit` passed to the API call is not supported by the active device."""
CUDA_ERROR_CONTEXT_ALREADY_IN_USE = 216
"""This indicates that the :class:`CUcontext` passed to the API call can only be bound to a single CPU thread at a time but is already bound to a CPU thread."""
CUDA_ERROR_PEER_ACCESS_UNSUPPORTED = 217
"""This indicates that peer access is not supported across the given devices."""
CUDA_ERROR_INVALID_PTX = 218
"""This indicates that a PTX JIT compilation failed."""
CUDA_ERROR_INVALID_GRAPHICS_CONTEXT = 219
"""This indicates an error with OpenGL or DirectX context."""
CUDA_ERROR_NVLINK_UNCORRECTABLE = 220
"""This indicates that an uncorrectable NVLink error was detected during the execution."""
CUDA_ERROR_JIT_COMPILER_NOT_FOUND = 221
"""This indicates that the PTX JIT compiler library was not found."""
CUDA_ERROR_UNSUPPORTED_PTX_VERSION = 222
"""This indicates that the provided PTX was compiled with an unsupported toolchain."""
CUDA_ERROR_JIT_COMPILATION_DISABLED = 223
"""This indicates that the PTX JIT compilation was disabled."""
CUDA_ERROR_UNSUPPORTED_EXEC_AFFINITY = 224
"""This indicates that the :class:`CUexecAffinityType` passed to the API call is not supported by the active device."""
CUDA_ERROR_INVALID_SOURCE = 300
"""This indicates that the device kernel source is invalid. This includes compilation / linker errors encountered in device code or user error."""
CUDA_ERROR_FILE_NOT_FOUND = 301
"""This indicates that the file specified was not found."""
CUDA_ERROR_SHARED_OBJECT_SYMBOL_NOT_FOUND = 302
"""This indicates that a link to a shared object failed to resolve."""
CUDA_ERROR_SHARED_OBJECT_INIT_FAILED = 303
"""This indicates that initialization of a shared object failed."""
CUDA_ERROR_OPERATING_SYSTEM = 304
"""This indicates that an OS call failed."""
CUDA_ERROR_INVALID_HANDLE = 400
"""This indicates that a resource handle passed to the API call was not valid. Resource handles are opaque types like :class:`CUstream` and :class:`CUevent`."""
CUDA_ERROR_ILLEGAL_STATE = 401
"""This indicates that a resource required by the API call is not in a valid state to perform the requested operation."""
CUDA_ERROR_NOT_FOUND = 500
"""This indicates that a named symbol was not found. Examples of symbols are global / constant variable names, driver function names, texture names, and surface names."""
CUDA_ERROR_NOT_READY = 600
"""This indicates that asynchronous operations issued previously have not completed yet. This result is not actually an error, but must be indicated differently than :data:`CUDA_SUCCESS` (which indicates completion). Calls that may return this value include :func:`cuEventQuery` and :func:`cuStreamQuery`."""
CUDA_ERROR_ILLEGAL_ADDRESS = 700
"""While executing a kernel, the device encountered a load or store instruction on an invalid memory address. This leaves the process in an inconsistent state and any further CUDA work will return the same error. To continue using CUDA, the process must be terminated and relaunched."""
CUDA_ERROR_LAUNCH_OUT_OF_RESOURCES = 701
"""This indicates that a launch did not occur because it did not have appropriate resources. This error usually indicates that the user has attempted to pass too many arguments to the device kernel, or the kernel launch specifies too many threads for the kernel's register count. Passing arguments of the wrong size (i.e. a 64 - bit pointer when a 32 - bit int is expected) is equivalent to passing too many arguments and can also result in this error."""
CUDA_ERROR_LAUNCH_TIMEOUT = 702
"""This indicates that the device kernel took too long to execute. This can only occur if timeouts are enabled - see the device attribute :data:`CU_DEVICE_ATTRIBUTE_KERNEL_EXEC_TIMEOUT` for more information. This leaves the process in an inconsistent state and any further CUDA work will return the same error. To continue using CUDA, the process must be terminated and relaunched."""
CUDA_ERROR_LAUNCH_INCOMPATIBLE_TEXTURING = 703
"""This error indicates a kernel launch that uses an incompatible texturing mode."""
CUDA_ERROR_PEER_ACCESS_ALREADY_ENABLED = 704
"""This error indicates that a call to :func:`cuCtxEnablePeerAccess` is trying to re - enable peer access to a context which has already had peer access to it enabled."""
CUDA_ERROR_PEER_ACCESS_NOT_ENABLED = 705
"""This error indicates that :func:`cuCtxDisablePeerAccess` is trying to disable peer access which has not been enabled yet via :func:`cuCtxEnablePeerAccess`."""
CUDA_ERROR_PRIMARY_CONTEXT_ACTIVE = 708
"""This error indicates that the primary context for the specified device has already been initialized."""
CUDA_ERROR_CONTEXT_IS_DESTROYED = 709
"""This error indicates that the context current to the calling thread has been destroyed using :func:`cuCtxDestroy`, or is a primary context which has not yet been initialized."""
CUDA_ERROR_ASSERT = 710
"""A device - side assert triggered during kernel execution. The context cannot be used anymore, and must be destroyed. All existing device memory allocations from this context are invalid and must be reconstructed if the program is to continue using CUDA."""
CUDA_ERROR_TOO_MANY_PEERS = 711
"""This error indicates that the hardware resources required to enable peer access have been exhausted for one or more of the devices passed to :func:`cuCtxEnablePeerAccess`."""
CUDA_ERROR_HOST_MEMORY_ALREADY_REGISTERED = 712
"""This error indicates that the memory range passed to :func:`cuMemHostRegister` has already been registered."""
CUDA_ERROR_HOST_MEMORY_NOT_REGISTERED = 713
"""This error indicates that the pointer passed to :func:`cuMemHostUnregister` does not correspond to any currently registered memory region."""
CUDA_ERROR_HARDWARE_STACK_ERROR = 714
"""While executing a kernel, the device encountered a stack error. This can be due to stack corruption or exceeding the stack size limit. This leaves the process in an inconsistent state and any further CUDA work will return the same error. To continue using CUDA, the process must be terminated and relaunched."""
CUDA_ERROR_ILLEGAL_INSTRUCTION = 715
"""While executing a kernel, the device encountered an illegal instruction. This leaves the process in an inconsistent state and any further CUDA work will return the same error. To continue using CUDA, the process must be terminated and relaunched."""
CUDA_ERROR_MISALIGNED_ADDRESS = 716
"""While executing a kernel, the device encountered a load or store instruction on a memory address which is not aligned. This leaves the process in an inconsistent state and any further CUDA work will return the same error. To continue using CUDA, the process must be terminated and relaunched."""
CUDA_ERROR_INVALID_ADDRESS_SPACE = 717
"""While executing a kernel, the device encountered an instruction which can only operate on memory locations in certain address spaces (global, shared, or local), but was supplied a memory address not belonging to an allowed address space. This leaves the process in an inconsistent state and any further CUDA work will return the same error. To continue using CUDA, the process must be terminated and relaunched."""
CUDA_ERROR_INVALID_PC = 718
"""While executing a kernel, the device program counter wrapped its address space. This leaves the process in an inconsistent state and any further CUDA work will return the same error. To continue using CUDA, the process must be terminated and relaunched."""
CUDA_ERROR_LAUNCH_FAILED = 719
"""An exception occurred on the device while executing a kernel. Common causes include dereferencing an invalid device pointer and accessing out of bounds shared memory. Less common cases can be system specific - more information about these cases can be found in the system specific user guide. This leaves the process in an inconsistent state and any further CUDA work will return the same error. To continue using CUDA, the process must be terminated and relaunched."""
CUDA_ERROR_COOPERATIVE_LAUNCH_TOO_LARGE = 720
"""This error indicates that the number of blocks launched per grid for a kernel that was launched via either :func:`cuLaunchCooperativeKernel` or :func:`cuLaunchCooperativeKernelMultiDevice` exceeds the maximum number of blocks as allowed by :func:`cuOccupancyMaxActiveBlocksPerMultiprocessor` or :func:`cuOccupancyMaxActiveBlocksPerMultiprocessorWithFlags` times the number of multiprocessors as specified by the device attribute :data:`CU_DEVICE_ATTRIBUTE_MULTIPROCESSOR_COUNT`."""
CUDA_ERROR_NOT_PERMITTED = 800
"""This error indicates that the attempted operation is not permitted."""
CUDA_ERROR_NOT_SUPPORTED = 801
"""This error indicates that the attempted operation is not supported on the current system or device."""
CUDA_ERROR_SYSTEM_NOT_READY = 802
"""This error indicates that the system is not yet ready to start any CUDA work. To continue using CUDA, verify the system configuration is in a valid state and all required driver daemons are actively running. More information about this error can be found in the system specific user guide."""
CUDA_ERROR_SYSTEM_DRIVER_MISMATCH = 803
"""This error indicates that there is a mismatch between the versions of the display driver and the CUDA driver. Refer to the compatibility documentation for supported versions."""
CUDA_ERROR_COMPAT_NOT_SUPPORTED_ON_DEVICE = 804
"""This error indicates that the system was upgraded to run with forward compatibility but the visible hardware detected by CUDA does not support this configuration. Refer to the compatibility documentation for the supported hardware matrix or ensure that only supported hardware is visible during initialization via the ``CUDA_VISIBLE_DEVICES`` environment variable."""
CUDA_ERROR_MPS_CONNECTION_FAILED = 805
"""This error indicates that the MPS client failed to connect to the MPS control daemon or the MPS server."""
CUDA_ERROR_MPS_RPC_FAILURE = 806
"""This error indicates that the remote procedural call between the MPS server and the MPS client failed."""
CUDA_ERROR_MPS_SERVER_NOT_READY = 807
"""This error indicates that the MPS server is not ready to accept new MPS client requests. This error can be returned when the MPS server is in the process of recovering from a fatal failure."""
CUDA_ERROR_MPS_MAX_CLIENTS_REACHED = 808
"""This error indicates that the hardware resources required to create MPS client have been exhausted."""
CUDA_ERROR_MPS_MAX_CONNECTIONS_REACHED = 809
"""This error indicates the the hardware resources required to support device connections have been exhausted."""
CUDA_ERROR_STREAM_CAPTURE_UNSUPPORTED = 900
"""This error indicates that the operation is not permitted when the stream is capturing."""
CUDA_ERROR_STREAM_CAPTURE_INVALIDATED = 901
"""This error indicates that the current capture sequence on the stream has been invalidated due to a previous error."""
CUDA_ERROR_STREAM_CAPTURE_MERGE = 902
"""This error indicates that the operation would have resulted in a merge of two independent capture sequences."""
CUDA_ERROR_STREAM_CAPTURE_UNMATCHED = 903
"""This error indicates that the capture was initiated not in this stream."""
CUDA_ERROR_STREAM_CAPTURE_UNJOINED = 904
"""This error indicates that the capture sequence contains a fork that was not joined to the primary stream."""
CUDA_ERROR_STREAM_CAPTURE_ISOLATION = 905
"""This error indicates that a dependency would have been created which crosses the capture sequence boundary. Only implicit in -stream ordering dependencies are allowed to cross the boundary."""
CUDA_ERROR_STREAM_CAPTURE_IMPLICIT = 906
"""This error indicates a disallowed implicit dependency on a current capture sequence from :func:`cudaStreamLegacy`."""
CUDA_ERROR_CAPTURED_EVENT = 907
"""This error indicates that the operation is not permitted on an event which was last recorded in a capturing stream."""
CUDA_ERROR_STREAM_CAPTURE_WRONG_THREAD = 908
"""A stream capture sequence not initiated with the :data:`CU_STREAM_CAPTURE_MODE_RELAXED` argument to :func:`cuStreamBeginCapture` was passed to :func:`cuStreamEndCapture` in a different thread."""
CUDA_ERROR_TIMEOUT = 909
"""This error indicates that the timeout specified for the wait operation has lapsed."""
CUDA_ERROR_GRAPH_EXEC_UPDATE_FAILURE = 910
"""This error indicates that the graph update was not performed because it included changes which violated constraints specific to instantiated graph update."""
CUDA_ERROR_EXTERNAL_DEVICE = 911
"""This indicates that an async error has occurred in a device outside of CUDA. If CUDA was waiting for an external device's signal before consuming shared data, the external device signaled an error indicating that the data is not valid for consumption. This leaves the process in an inconsistent state and any further CUDA work will return the same error. To continue using CUDA, the process must be terminated and relaunched."""
CUDA_ERROR_UNKNOWN = 999
"""This indicates that an unknown internal error has occurred."""
# pylint: enable=line-too-long


# Error Checking #
class CUDAError(Exception):
    """Base exception class for CUDA driver query errors."""

    _value_class_mapping: _ClassVar[dict[int, type[CUDAError]]] = {}
    _errcode_to_string: _ClassVar[dict[int, str]] = {  # List of currently known error codes
        CUDA_ERROR_NOT_INITIALIZED:                'Initialization error.',
        CUDA_ERROR_NOT_FOUND:                      'Named symbol not found.',
        CUDA_ERROR_INVALID_VALUE:                  'Invalid argument.',
        CUDA_ERROR_NO_DEVICE:                      'No CUDA-capable device is detected.',
        CUDA_ERROR_INVALID_DEVICE:                 'Invalid device ordinal.',
        CUDA_ERROR_SYSTEM_DRIVER_MISMATCH:         'System has unsupported display driver / CUDA driver combination.',
        CUDA_ERROR_DEINITIALIZED:                  'Driver shutting down.',
        CUDA_ERROR_COMPAT_NOT_SUPPORTED_ON_DEVICE: 'Forward compatibility was attempted on non supported Hardware.',
        CUDA_ERROR_INVALID_CONTEXT:                'Invalid device context.',
    }  # fmt:skip
    _errcode_to_name: _ClassVar[dict[int, str]] = {}
    value: int

    def __new__(cls, value: int) -> _Self:
        """Map value to a proper subclass of :class:`CUDAError`."""
        if cls is CUDAError:
            # pylint: disable-next=self-cls-assignment
            cls = CUDAError._value_class_mapping.get(value, cls)  # type: ignore[assignment]
        obj = Exception.__new__(cls)
        obj.value = value
        return obj

    def __repr__(self) -> str:
        """Return a string representation of the error."""
        # pylint: disable=no-member
        try:
            if self.value not in CUDAError._errcode_to_string:
                CUDAError._errcode_to_string[self.value] = '{}.'.format(
                    cuGetErrorString(self.value).rstrip('.').capitalize(),
                )
            if self.value not in CUDAError._errcode_to_name:
                CUDAError._errcode_to_name[self.value] = cuGetErrorName(self.value)
            return '{} Code: {} ({}).'.format(
                CUDAError._errcode_to_string[self.value],
                CUDAError._errcode_to_name[self.value],
                self.value,
            )
        except CUDAError:
            return f'CUDA Error with code {self.value}.'

    def __eq__(self, other: object) -> bool:
        """Test equality to other object."""
        if not isinstance(other, CUDAError):
            return NotImplemented
        return self.value == other.value  # pylint: disable=no-member

    def __reduce__(self) -> tuple[type[CUDAError], tuple[int]]:
        """Return state information for pickling."""
        return CUDAError, (self.value,)  # pylint: disable=no-member


def cudaExceptionClass(cudaErrorCode: int) -> type[CUDAError]:
    """Map value to a proper subclass of :class:`CUDAError`.

    Raises:
        ValueError: If the error code is not valid.
    """
    if cudaErrorCode not in CUDAError._value_class_mapping:  # pylint: disable=protected-access
        raise ValueError(f'cudaErrorCode {cudaErrorCode} is not valid.')
    return CUDAError._value_class_mapping[cudaErrorCode]  # pylint: disable=protected-access


def _extract_cuda_errors_as_classes() -> None:
    """Generate a hierarchy of classes on top of :class:`CUDAError` class.

    Each CUDA Error gets a new :class:`CUDAError` subclass. This way try-except blocks can filter
    appropriate exceptions more easily.

    :class:`CUDAError` is a parent class. Each ``CUDA_ERROR_*`` gets it's own subclass.
    e.g. :data:`CUDA_ERROR_INVALID_VALUE` will be turned into :class:`CUDAError_InvalidValue`.
    """
    this_module = _sys.modules[__name__]
    cuda_error_names = [x for x in dir(this_module) if x.startswith('CUDA_ERROR_')]
    for err_name in cuda_error_names:
        # e.g. Turn CUDA_ERROR_INVALID_VALUE into CUDAError_InvalidValue
        pascal_case = _string.capwords(err_name.replace('CUDA_ERROR_', ''), '_').replace('_', '')
        class_name = f'CUDAError_{pascal_case}'
        err_val = getattr(this_module, err_name)

        def gen_new(value: int) -> _Callable[[type[CUDAError]], CUDAError]:
            def new(cls: type[CUDAError]) -> CUDAError:
                return CUDAError.__new__(cls, value)

            return new

        # pylint: disable=protected-access
        new_error_class = type(class_name, (CUDAError,), {'__new__': gen_new(err_val)})
        new_error_class.__module__ = __name__
        if err_val in CUDAError._errcode_to_string:
            new_error_class.__doc__ = 'CUDA Error: {} Code: :data:`{}` ({}).'.format(
                CUDAError._errcode_to_string[err_val],
                err_name,
                err_val,
            )
        else:
            new_error_class.__doc__ = f'CUDA Error with code :data:`{err_name}` ({err_val})'
        setattr(this_module, class_name, new_error_class)
        CUDAError._value_class_mapping[err_val] = new_error_class
        CUDAError._errcode_to_name[err_val] = err_name


# Add explicit references to appease linters
class __CUDAError(CUDAError):
    value: int

    def __new__(cls) -> CUDAError:  # type: ignore[misc,empty-body]
        ...


CUDAError_NotInitialized: type[__CUDAError]
CUDAError_NotFound: type[__CUDAError]
CUDAError_InvalidValue: type[__CUDAError]
CUDAError_NoDevice: type[__CUDAError]
CUDAError_InvalidDevice: type[__CUDAError]
CUDAError_SystemDriverMismatch: type[__CUDAError]
CUDAError_Deinitialized: type[__CUDAError]
CUDAError_CompatNotSupportedOnDevice: type[__CUDAError]
CUDAError_InvalidContext: type[__CUDAError]

_extract_cuda_errors_as_classes()
del _extract_cuda_errors_as_classes


def _cudaCheckReturn(ret: _Any) -> _Any:
    if ret != CUDA_SUCCESS:
        raise CUDAError(ret)
    return ret


# Function access #
__cudaLib: _ctypes.CDLL | None = None
__initialized: bool = False
__libLoadLock: _threading.Lock = _threading.Lock()
# Function pointers are cached to prevent unnecessary libLoadLock locking
__cudaGetFunctionPointer_cache: dict[str, _ctypes._CFuncPtr] = {}  # type: ignore[name-defined]


def __cudaGetFunctionPointer(name: str) -> _ctypes._CFuncPtr:  # type: ignore[name-defined]
    """Get the function pointer from the CUDA driver library.

    Raises:
        CUDAError_NotInitialized:
            If cannot found the CUDA driver library.
        CUDAError_NotFound:
            If cannot found the function pointer.
    """
    if name in __cudaGetFunctionPointer_cache:
        return __cudaGetFunctionPointer_cache[name]

    with __libLoadLock:
        # Ensure library was loaded
        if __cudaLib is None:
            raise CUDAError(CUDA_ERROR_NOT_INITIALIZED)
        try:
            __cudaGetFunctionPointer_cache[name] = getattr(__cudaLib, name)
            return __cudaGetFunctionPointer_cache[name]
        except AttributeError as ex:
            raise CUDAError(CUDA_ERROR_NOT_FOUND) from ex


def __LoadCudaLibrary() -> None:
    """Load the library if it isn't loaded already.

    Raises:
        CUDAError_NotInitialized:
            If cannot found the CUDA driver library.
    """
    global __cudaLib  # pylint: disable=global-statement

    if __cudaLib is None:
        # Lock to ensure only one caller loads the library
        with __libLoadLock:
            # Ensure the library still isn't loaded
            if __cudaLib is None:
                # Platform specific libcuda location
                system = _platform.system()
                if system == 'Darwin':
                    lib_filenames = [
                        'libcuda.1.dylib',  # check library path first
                        'libcuda.dylib',
                        '/usr/local/cuda/lib/libcuda.1.dylib',
                        '/usr/local/cuda/lib/libcuda.dylib',
                    ]
                elif system == 'Linux':
                    lib_filenames = [
                        'libcuda.so',  # check library path first
                        '/usr/lib64/nvidia/libcuda.so',  # Redhat/CentOS/Fedora
                        '/usr/lib/x86_64-linux-gnu/libcuda.so',  # Ubuntu
                        '/usr/lib/wsl/lib/libcuda.so',  # WSL
                    ]
                    # Also add libraries with version suffix `.1`
                    lib_filenames = list(
                        _itertools.chain.from_iterable((f'{lib}.1', lib) for lib in lib_filenames),
                    )
                elif system == 'Windows':
                    bits = _platform.architecture()[0].replace('bit', '')  # e.g., '64' or '32'
                    lib_filenames = [f'nvcuda{bits}.dll', 'nvcuda.dll']
                # Open library
                for lib_filename in lib_filenames:
                    try:
                        __cudaLib = _ctypes.CDLL(lib_filename)
                        break
                    except OSError:
                        pass
                if __cudaLib is None:
                    _cudaCheckReturn(CUDA_ERROR_NOT_INITIALIZED)


def cuInit(flags: int = 0) -> None:
    """Initialize the CUDA driver API.

    Initialize the driver API and must be called before any other function from the driver API.
    Currently, the ``flags`` parameter must be :data:`0`. If :func:`cuInit` has not been called,
    any function from the driver API will return :data:`CUDA_ERROR_NOT_INITIALIZED`.

    Raises:
        CUDAError_NoDevice:
            If no CUDA-capable device is available.
        CUDAError_InvalidDevice:
            If the device ordinal supplied by the user does not correspond to a valid CUDA device or
            that the action requested is invalid for the specified device.
        CUDAError_SystemDriverMismatch:
            If there is a mismatch between the versions of the display driver and the CUDA driver.
        CUDAError_CompatNotSupportedOnDevice:
            If the system was upgraded to run with forward compatibility but the visible hardware
            detected by CUDA does not support this configuration.
        CUDAError_InvalidValue:
            If passed with invalid flag value.
        CUDAError_NotInitialized:
            If cannot found the CUDA driver library.
    """
    global __initialized  # pylint: disable=global-statement

    __LoadCudaLibrary()

    if __initialized:
        return

    fn = __cudaGetFunctionPointer('cuInit')

    ret = fn(_ctypes.c_uint(flags))
    _cudaCheckReturn(ret)

    with __libLoadLock:
        __initialized = True


def cuGetErrorName(error: int) -> str:
    """Get the string representation of an error code enum name.

    Raises:
        CUDAError_InvalidValue:
            If the error code is not recognized.
        CUDAError_NotInitialized:
            If the CUDA driver API is not initialized.
    """
    fn = __cudaGetFunctionPointer('cuGetErrorName')

    p_name = _ctypes.POINTER(_ctypes.c_char_p)()
    ret = fn(_CUresult_t(error), _ctypes.byref(p_name))
    _cudaCheckReturn(ret)
    name = _ctypes.string_at(p_name)
    return name.decode('utf-8', errors='replace')


def cuGetErrorString(error: int) -> str:
    """Get the string description of an error code.

    Raises:
        CUDAError_InvalidValue:
            If the error code is not recognized.
        CUDAError_NotInitialized:
            If the CUDA driver API is not initialized.
    """
    fn = __cudaGetFunctionPointer('cuGetErrorString')

    p_name = _ctypes.POINTER(_ctypes.c_char_p)()
    ret = fn(_CUresult_t(error), _ctypes.byref(p_name))
    _cudaCheckReturn(ret)
    name = _ctypes.string_at(p_name)
    return name.decode('utf-8', errors='replace')


def cuDriverGetVersion() -> str:
    """Get the latest CUDA version supported by driver.

    Returns:
        A string of the form :data:`'<major>.<minor>'`.

    Raises:
        CUDAError_InvalidValue:
            If the driver call fails.
        CUDAError_NotInitialized:
            If the CUDA driver API is not initialized.
    """
    fn = __cudaGetFunctionPointer('cuDriverGetVersion')

    driver_version = _ctypes.c_int()
    ret = fn(_ctypes.byref(driver_version))
    _cudaCheckReturn(ret)
    major = driver_version.value // 1000
    minor = (driver_version.value % 1000) // 10
    return f'{major}.{minor}'


def cuDeviceGetCount() -> int:
    """Get the number of compute-capable devices.

    Returns: int
        The number of devices with compute capability greater than or equal to 2.0 that are available
        for execution. If there is no such device, :func:`cuDeviceGetCount` returns :data:`0`.

    Raises:
        CUDAError_InvalidContext:
            If there is no context bound to the current thread.
        CUDAError_InvalidValue:
            If the driver call fails.
        CUDAError_Deinitialized:
            If the CUDA driver in the process is shutting down.
        CUDAError_NotInitialized:
            If the CUDA driver API is not initialized.
    """
    fn = __cudaGetFunctionPointer('cuDeviceGetCount')

    count = _ctypes.c_int(0)
    ret = fn(_ctypes.byref(count))
    _cudaCheckReturn(ret)
    return count.value


def cuDeviceGet(ordinal: int) -> _c_CUdevice_t:
    """Get a handle to a compute device.

    Returns:
        A device handle given an ordinal in the range :code:`[0, ..., cuDeviceGetCount() - 1]`.

    Raises:
        CUDAError_InvalidContext:
            If there is no context bound to the current thread.
        CUDAError_InvalidDevice:
            If the device ordinal supplied by the user does not correspond to a valid CUDA device or
            that the action requested is invalid for the specified device.
        CUDAError_InvalidValue:
            If the driver call fails.
        CUDAError_Deinitialized:
            If the CUDA driver in the process is shutting down.
        CUDAError_NotInitialized:
            If the CUDA driver API is not initialized.
    """
    fn = __cudaGetFunctionPointer('cuDeviceGet')

    device = _c_CUdevice_t()
    ret = fn(_ctypes.byref(device), _ctypes.c_int(ordinal))
    _cudaCheckReturn(ret)
    return device


def cuDeviceGetByPCIBusId(pciBusId: str) -> _c_CUdevice_t:
    """Get a handle to a compute device.

    Args:
        pciBusId (str):
            String in one of the following forms: ``[domain]:[bus]:[device].[function]``,
            ``[domain]:[bus]:[device]``, ``[bus]:[device].[function]`` where ``domain``, ``bus``,
            ``device``, and ``function`` are all hexadecimal values.

    Returns: int
        A device handle given a PCI bus ID string.

    Raises:
        CUDAError_InvalidDevice:
            If the device ordinal supplied by the user does not correspond to a valid CUDA device or
            that the action requested is invalid for the specified device.
        CUDAError_InvalidValue:
            If the value of :data:`pciBusId` is not a valid PCI bus identifier.
        CUDAError_Deinitialized:
            If the CUDA driver in the process is shutting down.
        CUDAError_NotInitialized:
            If the CUDA driver API is not initialized.
    """
    fn = __cudaGetFunctionPointer('cuDeviceGetByPCIBusId')

    device = _c_CUdevice_t()
    ret = fn(_ctypes.byref(device), _ctypes.c_char_p(pciBusId.encode('utf-8')))
    _cudaCheckReturn(ret)
    return device


def cuDeviceGetPCIBusId(device: _c_CUdevice_t) -> str:
    """Get a PCI Bus Id string for the device.

    Returns: str
        An identifier string for the device in the following format ``[domain]:[bus]:[device].[function]``
        where ``domain``, ``bus``, ``device``, and ``function`` are all hexadecimal values.

    Raises:
        CUDAError_InvalidDevice:
            If the device ordinal supplied by the user does not correspond to a valid CUDA device or
            that the action requested is invalid for the specified device.
        CUDAError_InvalidValue:
            If the driver call fails.
        CUDAError_Deinitialized:
            If the CUDA driver in the process is shutting down.
        CUDAError_NotInitialized:
            If the CUDA driver API is not initialized.
    """
    fn = __cudaGetFunctionPointer('cuDeviceGetPCIBusId')

    pciBusId = _ctypes.create_string_buffer(256)
    ret = fn(pciBusId, _ctypes.c_int(256), device)
    _cudaCheckReturn(ret)
    return pciBusId.value.decode('utf-8', errors='replace')


def cuDeviceGetName(device: _c_CUdevice_t) -> str:
    """Get an identifier string for the device.

    Returns: str
        An ASCII string identifying the device.

    Raises:
        CUDAError_InvalidContext:
            If there is no context bound to the current thread.
        CUDAError_InvalidDevice:
            If the device ordinal supplied by the user does not correspond to a valid CUDA device or
            that the action requested is invalid for the specified device.
        CUDAError_InvalidValue:
            If the driver call fails.
        CUDAError_Deinitialized:
            If the CUDA driver in the process is shutting down.
        CUDAError_NotInitialized:
            If the CUDA driver API is not initialized.
    """
    fn = __cudaGetFunctionPointer('cuDeviceGetName')

    name = _ctypes.create_string_buffer(256)
    ret = fn(name, _ctypes.c_int(256), device)
    _cudaCheckReturn(ret)
    return name.value.decode('utf-8', errors='replace')


def cuDeviceGetUuid(device: _c_CUdevice_t) -> str:
    """Get a UUID for the device.

    Raises:
        CUDAError_InvalidDevice:
            If the device ordinal supplied by the user does not correspond to a valid CUDA device or
            that the action requested is invalid for the specified device.
        CUDAError_InvalidValue:
            If the driver call fails.
        CUDAError_Deinitialized:
            If the CUDA driver in the process is shutting down.
        CUDAError_NotInitialized:
            If the CUDA driver API is not initialized.
    """
    try:
        fn = __cudaGetFunctionPointer('cuDeviceGetUuid_v2')
    except CUDAError_NotFound:  # noqa: F821 # pylint: disable=undefined-variable
        fn = __cudaGetFunctionPointer('cuDeviceGetUuid')

    uuid = (_ctypes.c_ubyte * 16)()
    ret = fn(uuid, device)
    _cudaCheckReturn(ret)
    uuid = ''.join(map('{:02x}'.format, uuid))
    return '-'.join((uuid[:8], uuid[8:12], uuid[12:16], uuid[16:20], uuid[20:32]))


def cuDeviceGetUuid_v2(device: _c_CUdevice_t) -> str:
    """Get a UUID for the device (CUDA 11.4+).

    Raises:
        CUDAError_InvalidDevice:
            If the device ordinal supplied by the user does not correspond to a valid CUDA device or
            that the action requested is invalid for the specified device.
        CUDAError_InvalidValue:
            If the driver call fails.
        CUDAError_Deinitialized:
            If the CUDA driver in the process is shutting down.
        CUDAError_NotInitialized:
            If the CUDA driver API is not initialized.
    """
    fn = __cudaGetFunctionPointer('cuDeviceGetUuid_v2')

    uuid = (_ctypes.c_ubyte * 16)()
    ret = fn(uuid, device)
    _cudaCheckReturn(ret)
    uuid = ''.join(map('{:02x}'.format, uuid))
    return '-'.join((uuid[:8], uuid[8:12], uuid[12:16], uuid[16:20], uuid[20:32]))


def cuDeviceTotalMem(device: _c_CUdevice_t) -> int:
    """Get the total amount of memory on the device (in bytes).

    Raises:
        CUDAError_InvalidContext:
            If there is no context bound to the current thread.
        CUDAError_InvalidDevice:
            If the device ordinal supplied by the user does not correspond to a valid CUDA device or
            that the action requested is invalid for the specified device.
        CUDAError_InvalidValue:
            If the driver call fails.
        CUDAError_Deinitialized:
            If the CUDA driver in the process is shutting down.
        CUDAError_NotInitialized:
            If the CUDA driver API is not initialized.
    """
    fn = __cudaGetFunctionPointer('cuDeviceTotalMem')

    bytes = _ctypes.c_size_t()  # pylint: disable=redefined-builtin
    ret = fn(_ctypes.byref(bytes), device)
    _cudaCheckReturn(ret)
    return bytes.value


def is_available() -> bool:
    """Test whether there are any CUDA visible devices."""
    try:
        return cuDeviceGetCount() > 0
    except CUDAError:
        return False



# This file is part of nvitop, the interactive NVIDIA-GPU process viewer.
# This file is originally part of ranger, the console file manager. https://github.com/ranger/ranger
# License: GNU GPL version 3.

# pylint: disable=missing-module-docstring,missing-class-docstring,missing-function-docstring

import curses


class MouseEvent:
    PRESSED = [
        0,
        curses.BUTTON1_PRESSED,
        curses.BUTTON2_PRESSED,
        curses.BUTTON3_PRESSED,
        curses.BUTTON4_PRESSED,
    ]
    RELEASED = [
        0,
        curses.BUTTON1_RELEASED,
        curses.BUTTON2_RELEASED,
        curses.BUTTON3_RELEASED,
        curses.BUTTON4_RELEASED,
    ]
    CLICKED = [
        0,
        curses.BUTTON1_CLICKED,
        curses.BUTTON2_CLICKED,
        curses.BUTTON3_CLICKED,
        curses.BUTTON4_CLICKED,
    ]
    DOUBLE_CLICKED = [
        0,
        curses.BUTTON1_DOUBLE_CLICKED,
        curses.BUTTON2_DOUBLE_CLICKED,
        curses.BUTTON3_DOUBLE_CLICKED,
        curses.BUTTON4_DOUBLE_CLICKED,
    ]
    CTRL_SCROLLWHEEL_MULTIPLIER = 5

    def __init__(self, state):
        """Create a MouseEvent object from the result of win.getmouse()."""
        _, self.x, self.y, _, self.bstate = state

        # x-values above ~220 suddenly became negative, apparently
        # it's sufficient to add 0xFF to fix that error.
        if self.x < 0:
            self.x += 0xFF

        if self.y < 0:
            self.y += 0xFF

    def pressed(self, n):
        """Return whether the mouse key n is pressed."""
        try:
            return (self.bstate & MouseEvent.PRESSED[n]) != 0
        except IndexError:
            return False

    def released(self, n):
        """Return whether the mouse key n is released."""
        try:
            return (self.bstate & MouseEvent.RELEASED[n]) != 0
        except IndexError:
            return False

    def clicked(self, n):
        """Return whether the mouse key n is clicked."""
        try:
            return (self.bstate & MouseEvent.CLICKED[n]) != 0
        except IndexError:
            return False

    def double_clicked(self, n):
        """Return whether the mouse key n is double clicked."""
        try:
            return (self.bstate & MouseEvent.DOUBLE_CLICKED[n]) != 0
        except IndexError:
            return False

    def wheel_direction(self):
        """Return the direction of the scroll action, 0 if there was none."""
        # If the bstate > ALL_MOUSE_EVENTS, it's an invalid mouse button.
        # I interpret invalid buttons as "scroll down" because all tested
        # systems have a broken curses implementation and this is a workaround.
        # Recently it seems to have been fixed, as 2**21 was introduced as
        # the code for the "scroll down" button.
        if self.pressed(4):
            return -self.CTRL_SCROLLWHEEL_MULTIPLIER if self.ctrl() else -1
        if self.pressed(2) or (self.bstate & (1 << 21)) or self.key_invalid():
            return self.CTRL_SCROLLWHEEL_MULTIPLIER if self.ctrl() else 1
        return 0

    def ctrl(self):
        return self.bstate & curses.BUTTON_CTRL

    def alt(self):
        return self.bstate & curses.BUTTON_ALT

    def shift(self):
        return self.bstate & curses.BUTTON_SHIFT

    def key_invalid(self):
        return self.bstate > curses.ALL_MOUSE_EVENTS



# This file is part of nvitop, the interactive NVIDIA-GPU process viewer.
#
# Copyright 2021-2023 Xuehai Pan. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Prometheus exporter built on top of ``nvitop``."""

__version__ = '1.3.2'
__license__ = 'Apache-2.0'
__author__ = __maintainer__ = 'Xuehai Pan'
__email__ = 'XuehaiPan@pku.edu.cn'
__release__ = False

if not __release__:
    import os
    import subprocess

    try:
        prefix, sep, suffix = (
            subprocess.check_output(
                ['git', 'describe', '--abbrev=7'],  # noqa: S603,S607
                cwd=os.path.dirname(os.path.abspath(__file__)),
                stderr=subprocess.DEVNULL,
                text=True,
            )
            .strip()
            .lstrip('v')
            .replace('-', '.dev', 1)
            .replace('-', '+', 1)
            .partition('.dev')
        )
        if sep:
            version_prefix, dot, version_tail = prefix.rpartition('.')
            prefix = f'{version_prefix}{dot}{int(version_tail) + 1}'
            __version__ = sep.join((prefix, suffix))
            del version_prefix, dot, version_tail
        else:
            __version__ = prefix
        del prefix, sep, suffix
    except (OSError, subprocess.CalledProcessError):
        pass

    del os, subprocess



# This file is part of nvitop, the interactive NVIDIA-GPU process viewer.
# License: GNU GPL version 3.

"""The interactive NVIDIA-GPU process viewer."""

import sys

from nvitop.cli import main


if __name__ == '__main__':
    sys.exit(main())



# This file is part of nvitop, the interactive NVIDIA-GPU process viewer.
# License: GNU GPL version 3.

# pylint: disable=missing-module-docstring

from nvitop.gui.screens.environ import EnvironScreen
from nvitop.gui.screens.help import HelpScreen
from nvitop.gui.screens.main import BreakLoop, MainScreen
from nvitop.gui.screens.metrics import ProcessMetricsScreen
from nvitop.gui.screens.treeview import TreeViewScreen



# This file is part of nvitop, the interactive NVIDIA-GPU process viewer.
# This file is originally part of ranger, the console file manager. https://github.com/ranger/ranger
# License: GNU GPL version 3.

# pylint: disable=missing-module-docstring,missing-function-docstring

from nvitop.gui.library.libcurses import CursesShortcuts


class Displayable(CursesShortcuts):  # pylint: disable=too-many-instance-attributes
    """Displayables are objects which are displayed on the screen.

    This is just the abstract class, defining basic operations
    such as resizing, printing, changing colors.
    Subclasses of displayable can extend these methods:

    draw() -- draw the object. Is only called if visible.
    poke() -- is called just before draw(), even if not visible.
    finalize() -- called after all objects finished drawing.
    press(key) -- called after a key press on focused objects.
    destroy() -- called before destroying the displayable object

    Additionally, there are these methods:

    __contains__(item) -- is the item (y, x) inside the panel?

    These attributes are set:

    Modifiable:
        focused -- Focused objects receive press() calls.
        visible -- Visible objects receive draw() and finalize() calls
        need_redraw -- Should the panel be redrawn? This variable may
            be set at various places in the script and should eventually be
            handled (and unset) in the draw() method.

    Read-Only: (i.e. recommended not to change manually)
        win -- the own curses window object
        parent -- the parent (DisplayableContainer) object or None
        x, y, width, height -- absolute coordinates and boundaries
    """

    def __init__(self, win, root=None):
        super().__init__()

        self._need_redraw = True
        self.focused = False
        self._old_visible = self._visible = True
        self.x = 0
        self.y = 0
        self._width = 0
        self.height = 0

        self.win = win
        self.root = root
        self.parent = None

    def __contains__(self, item):
        """Check if item is inside the boundaries.

        item can be an iterable like [y, x] or an object with x and y methods.
        """
        try:
            y, x = item.y, item.x
        except AttributeError:
            try:
                y, x = item
            except (ValueError, TypeError):
                return False

        return self.contains_point(y, x)

    def contains_point(self, y, x):
        """Test whether the point lies inside this object.

        x and y should be absolute coordinates.
        """
        return (self.x <= x < self.x + self.width) and (self.y <= y < self.y + self.height)

    def poke(self):
        """Called before drawing, even if invisible."""
        if self._old_visible != self.visible:
            self._old_visible = self.visible
            self.need_redraw = True

            if not self.visible:
                self.win.erase()

    def draw(self):
        """Draw the object.

        Called on every main iteration if visible. Containers should call draw()
        on their contained objects here. Override this!
        """
        self.need_redraw = False

    def finalize(self):
        """Called after every displayable is done drawing.

        Override this!
        """
        self.need_redraw = False

    def destroy(self):
        """Called when the object is destroyed."""
        self.win = None
        self.root = None

    def click(self, event):
        """Called when a mouse key is pressed and self.focused is True.

        Override this!
        """

    def press(self, key):
        """Called when a key is pressed and self.focused is True.

        Override this!
        """

    @property
    def visible(self):
        return self._visible

    @visible.setter
    def visible(self, value):
        if self._visible != value:
            self.need_redraw = True
            self._visible = value
        if not self.visible:
            self.focused = False

    @property
    def need_redraw(self):
        return self._need_redraw

    @need_redraw.setter
    def need_redraw(self, value):
        if self._need_redraw != value:
            self._need_redraw = value
            if value and self.parent is not None and not self.parent.need_redraw:
                self.parent.need_redraw = True

    @property
    def width(self):
        return self._width

    @width.setter
    def width(self, value):
        if self.width != value and self.visible:
            self.need_redraw = True
        self._width = value

    def __str__(self):
        return self.__class__.__name__


class DisplayableContainer(Displayable):
    """DisplayableContainers are Displayables which contain other Displayables.

    This is also an abstract class. The methods draw, poke, finalize,
    click, press and destroy are extended here and will recursively
    call the function on all contained objects.

    New methods:

    add_child(object) -- add the object to the container.
    replace_child(old_obj, new_obj) -- replaces old object with new object.
    remove_child(object) -- remove the object from the container.

    New attributes:

    container -- a list with all contained objects (rw)
    """

    def __init__(self, win, root=None):
        super().__init__(win, root)

        self.container = []

    # extended or overridden methods

    def poke(self):
        """Recursively called on objects in container."""
        super().poke()
        for displayable in self.container:
            displayable.poke()

    def draw(self):
        """Recursively called on visible objects in container."""
        for displayable in self.container:
            if self.need_redraw:
                displayable.need_redraw = True
            if displayable.visible:
                displayable.draw()

        self.need_redraw = False

    def finalize(self):
        """Recursively called on visible objects in container."""
        for displayable in self.container:
            if displayable.visible:
                displayable.finalize()

    def destroy(self):
        """Recursively called on objects in container."""
        for displayable in self.container:
            displayable.destroy()
        super().destroy()

    def press(self, key):
        """Recursively called on objects in container."""
        focused_obj = self.get_focused_obj()

        if focused_obj:
            focused_obj.press(key)
            return True
        return False

    def click(self, event):
        """Recursively called on objects in container."""
        focused_obj = self.get_focused_obj()
        if focused_obj and focused_obj.click(event):
            return True

        return any(
            displayable.visible and event in displayable and displayable.click(event)
            for displayable in self.container
        )

    # new methods

    def add_child(self, obj):
        """Add the objects to the container."""
        if obj.parent is not None:
            obj.parent.remove_child(obj)
        self.container.append(obj)
        obj.parent = self
        obj.root = self.root

    def replace_child(self, old_obj, new_obj):
        """Replace the old object with the new instance in the container."""
        self.container[self.container.index(old_obj)] = new_obj
        new_obj.parent = self
        new_obj.root = self.root

    def remove_child(self, obj):
        """Remove the object from the container."""
        try:
            self.container.remove(obj)
        except ValueError:
            pass
        else:
            obj.parent = None
            obj.root = None

    def get_focused_obj(self):
        # Finds a focused displayable object in the container.
        for displayable in self.container:
            if displayable.focused:
                return displayable
            try:
                obj = displayable.get_focused_obj()
            except AttributeError:
                pass
            else:
                if obj is not None:
                    return obj
        return None



# This file is part of nvitop, the interactive NVIDIA-GPU process viewer.
#
# Copyright 2021-2023 Xuehai Pan. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Python bindings for the `CUDA Runtime APIs <https://docs.nvidia.com/cuda/cuda-runtime-api>`_."""

# pylint: disable=invalid-name

from __future__ import annotations

import ctypes as _ctypes
import glob as _glob
import os as _os
import platform as _platform
import sys as _sys
import threading as _threading
from typing import TYPE_CHECKING as _TYPE_CHECKING
from typing import Any as _Any
from typing import Callable as _Callable
from typing import ClassVar as _ClassVar


if _TYPE_CHECKING:
    from typing_extensions import Self as _Self  # Python 3.11+


_cudaError_t = _ctypes.c_int

# Error codes #
# pylint: disable=line-too-long
cudaSuccess = 0
"""The API call returned with no errors. In the case of query calls, this also means that the operation being queried is complete (see :func:`cudaEventQuery` and :func:`cudaStreamQuery`)."""
cudaErrorInvalidValue = 1
"""This indicates that one or more of the parameters passed to the API call is not within an acceptable range of values."""
cudaErrorMemoryAllocation = 2
"""The API call failed because it was unable to allocate enough memory to perform the requested operation."""
cudaErrorInitializationError = 3
"""The API call failed because the CUDA driver and runtime could not be initialized."""
cudaErrorCudartUnloading = 4
"""This indicates that a CUDA Runtime API call cannot be executed because it is being called during process shut down, at a point in time after CUDA driver has been unloaded."""
cudaErrorProfilerDisabled = 5
"""This indicates profiler is not initialized for this run. This can happen when the application is running with external profiling tools like visual profiler."""
cudaErrorInvalidConfiguration = 9
"""This indicates that a kernel launch is requesting resources that can never be satisfied by the current device. Requesting more shared memory per block than the device supports will trigger this error, as will requesting too many threads or blocks. See cudaDeviceProp for more device limitations."""
cudaErrorInvalidPitchValue = 12
"""This indicates that one or more of the pitch-related parameters passed to the API call is not within the acceptable range for pitch."""
cudaErrorInvalidSymbol = 13
"""This indicates that the symbol name / identifier passed to the API call is not a valid name or identifier."""
cudaErrorInvalidTexture = 18
"""This indicates that the texture passed to the API call is not a valid texture."""
cudaErrorInvalidTextureBinding = 19
"""This indicates that the texture binding is not valid. This occurs if you call :func:`cudaGetTextureAlignmentOffset` with an unbound texture."""
cudaErrorInvalidChannelDescriptor = 20
"""This indicates that the channel descriptor passed to the API call is not valid. This occurs if the format is not one of the formats specified by :data:`cudaChannelFormatKind`, or if one of the dimensions is invalid."""
cudaErrorInvalidMemcpyDirection = 21
"""This indicates that the direction of the :func:`memcpy` passed to the API call is not one of the types specified by :data:`cudaMemcpyKind`."""
cudaErrorInvalidFilterSetting = 26
"""This indicates that a non-float texture was being accessed with linear filtering. This is not supported by CUDA."""
cudaErrorInvalidNormSetting = 27
"""This indicates that an attempt was made to read a non-float texture as a normalized float. This is not supported by CUDA."""
cudaErrorStubLibrary = 34
"""This indicates that the CUDA driver that the application has loaded is a stub library. Applications that run with the stub rather than a real driver loaded will result in CUDA API returning this error."""
cudaErrorInsufficientDriver = 35
"""This indicates that the installed NVIDIA CUDA driver is older than the CUDA Runtime library. This is not a supported configuration. Users should install an updated NVIDIA display driver to allow the application to run."""
cudaErrorCallRequiresNewerDriver = 36
"""This indicates that the API call requires a newer CUDA driver than the one currently installed. Users should install an updated NVIDIA CUDA driver to allow the API call to succeed."""
cudaErrorInvalidSurface = 37
"""This indicates that the surface passed to the API call is not a valid surface."""
cudaErrorDuplicateVariableName = 43
"""This indicates that multiple global or constant variables (across separate CUDA source files in the application) share the same string name."""
cudaErrorDuplicateTextureName = 44
"""This indicates that multiple textures (across separate CUDA source files in the application) share the same string name."""
cudaErrorDuplicateSurfaceName = 45
"""This indicates that multiple surfaces (across separate CUDA source files in the application) share the same string name."""
cudaErrorDevicesUnavailable = 46
"""This indicates that all CUDA devices are busy or unavailable at the current time. Devices are often busy / unavailable due to use of :data:`cudaComputeModeProhibited`, :data:`cudaComputeModeExclusiveProcess`, or when long running CUDA kernels have filled up the GPU and are blocking new work from starting. They can also be unavailable due to memory constraints on a device that already has active CUDA work being performed."""
cudaErrorIncompatibleDriverContext = 49
"""This indicates that the current context is not compatible with this the CUDA Runtime. This can only occur if you are using CUDA Runtime / Driver interoperability and have created an existing Driver context using the driver API. The Driver context may be incompatible either because the Driver context was created using an older version of the API, because the Runtime API call expects a primary driver context and the Driver context is not primary, or because the Driver context has been destroyed."""
cudaErrorMissingConfiguration = 52
"""The device function being invoked (usually via :func:`cudaLaunchKernel`) was not previously configured via the :func:`cudaConfigureCall` function."""
cudaErrorLaunchMaxDepthExceeded = 65
"""This error indicates that a device runtime grid launch did not occur because the depth of the child grid would exceed the maximum supported number of nested grid launches."""
cudaErrorLaunchFileScopedTex = 66
"""This error indicates that a grid launch did not occur because the kernel uses file-scoped textures which are unsupported by the device runtime. Kernels launched via the device runtime only support textures created with the Texture Object API's."""
cudaErrorLaunchFileScopedSurf = 67
"""This error indicates that a grid launch did not occur because the kernel uses file-scoped surfaces which are unsupported by the device runtime. Kernels launched via the device runtime only support surfaces created with the Surface Object API's."""
cudaErrorSyncDepthExceeded = 68
"""This error indicates that a call to :func:`cudaDeviceSynchronize` made from the device runtime failed because the call was made at grid depth greater than than either the default (2 levels of grids) or user specified device limit :data:`cudaLimitDevRuntimeSyncDepth`. To be able to synchronize on launched grids at a greater depth successfully, the maximum nested depth at which :func:`cudaDeviceSynchronize` will be called must be specified with the :data:`cudaLimitDevRuntimeSyncDepth` limit to the :func:`cudaDeviceSetLimit` api before the host-side launch of a kernel using the device runtime. Keep in mind that additional levels of sync depth require the runtime to reserve large amounts of device memory that cannot be used for user allocations. Note that :func:`cudaDeviceSynchronize` made from device runtime is only supported on devices of compute capability < 9.0."""
cudaErrorLaunchPendingCountExceeded = 69
"""This error indicates that a device runtime grid launch failed because the launch would exceed the limit :data:`cudaLimitDevRuntimePendingLaunchCount`. For this launch to proceed successfully, :func:`cudaDeviceSetLimit` must be called to set the :data:`cudaLimitDevRuntimePendingLaunchCount` to be higher than the upper bound of outstanding launches that can be issued to the device runtime. Keep in mind that raising the limit of pending device runtime launches will require the runtime to reserve device memory that cannot be used for user allocations."""
cudaErrorInvalidDeviceFunction = 98
"""The requested device function does not exist or is not compiled for the proper device architecture."""
cudaErrorNoDevice = 100
"""This indicates that no CUDA-capable devices were detected by the installed CUDA driver."""
cudaErrorInvalidDevice = 101
"""This indicates that the device ordinal supplied by the user does not correspond to a valid CUDA device or that the action requested is invalid for the specified device."""
cudaErrorDeviceNotLicensed = 102
"""This indicates that the device doesn't have a valid Grid License."""
cudaErrorSoftwareValidityNotEstablished = 103
"""By default, the CUDA Runtime may perform a minimal set of self-tests, as well as CUDA driver tests, to establish the validity of both. Introduced in CUDA 11.2, this error return indicates that at least one of these tests has failed and the validity of either the runtime or the driver could not be established."""
cudaErrorStartupFailure = 127
"""This indicates an internal startup failure in the CUDA Runtime."""
cudaErrorInvalidKernelImage = 200
"""This indicates that the device kernel image is invalid."""
cudaErrorDeviceUninitialized = 201
"""This most frequently indicates that there is no context bound to the current thread. This can also be returned if the context passed to an API call is not a valid handle (such as a context that has had :func`cuCtxDestroy` invoked on it). This can also be returned if a user mixes different API versions (i.e. 3010 context with 3020 API calls)."""
cudaErrorMapBufferObjectFailed = 205
"""This indicates that the buffer object could not be mapped."""
cudaErrorUnmapBufferObjectFailed = 206
"""This indicates that the buffer object could not be unmapped."""
cudaErrorArrayIsMapped = 207
"""This indicates that the specified array is currently mapped and thus cannot be destroyed."""
cudaErrorAlreadyMapped = 208
"""This indicates that the resource is already mapped."""
cudaErrorNoKernelImageForDevice = 209
"""This indicates that there is no kernel image available that is suitable for the device. This can occur when a user specifies code generation options for a particular CUDA source file that do not include the corresponding device configuration."""
cudaErrorAlreadyAcquired = 210
"""This indicates that a resource has already been acquired."""
cudaErrorNotMapped = 211
"""This indicates that a resource is not mapped."""
cudaErrorNotMappedAsArray = 212
"""This indicates that a mapped resource is not available for access as an array."""
cudaErrorNotMappedAsPointer = 213
"""This indicates that a mapped resource is not available for access as a pointer."""
cudaErrorECCUncorrectable = 214
"""This indicates that an uncorrectable ECC error was detected during execution."""
cudaErrorUnsupportedLimit = 215
"""This indicates that the :class:`cudaLimit` passed to the API call is not supported by the active device."""
cudaErrorDeviceAlreadyInUse = 216
"""This indicates that a call tried to access an exclusive-thread device that is already in use by a different thread."""
cudaErrorPeerAccessUnsupported = 217
"""This error indicates that P2P access is not supported across the given devices."""
cudaErrorInvalidPtx = 218
"""A PTX compilation failed. The runtime may fall back to compiling PTX if an application does not contain a suitable binary for the current device."""
cudaErrorInvalidGraphicsContext = 219
"""This indicates an error with the OpenGL or DirectX context."""
cudaErrorNvlinkUncorrectable = 220
"""This indicates that an uncorrectable NVLink error was detected during the execution."""
cudaErrorJitCompilerNotFound = 221
"""This indicates that the PTX JIT compiler library was not found. The JIT Compiler library is used for PTX compilation. The runtime may fall back to compiling PTX if an application does not contain a suitable binary for the current device."""
cudaErrorUnsupportedPtxVersion = 222
"""This indicates that the provided PTX was compiled with an unsupported toolchain. The most common reason for this, is the PTX was generated by a compiler newer than what is supported by the CUDA driver and PTX JIT compiler."""
cudaErrorJitCompilationDisabled = 223
"""This indicates that the JIT compilation was disabled. The JIT compilation compiles PTX. The runtime may fall back to compiling PTX if an application does not contain a suitable binary for the current device."""
cudaErrorUnsupportedExecAffinity = 224
"""This indicates that the provided execution affinity is not supported by the device."""
cudaErrorInvalidSource = 300
"""This indicates that the device kernel source is invalid."""
cudaErrorFileNotFound = 301
"""This indicates that the file specified was not found."""
cudaErrorSharedObjectSymbolNotFound = 302
"""This indicates that a link to a shared object failed to resolve."""
cudaErrorSharedObjectInitFailed = 303
"""This indicates that initialization of a shared object failed."""
cudaErrorOperatingSystem = 304
"""This error indicates that an OS call failed."""
cudaErrorInvalidResourceHandle = 400
"""This indicates that a resource handle passed to the API call was not valid. Resource handles are opaque types like :data:`cudaStream_t` and :data:`cudaEvent_t`."""
cudaErrorIllegalState = 401
"""This indicates that a resource required by the API call is not in a valid state to perform the requested operation."""
cudaErrorSymbolNotFound = 500
"""This indicates that a named symbol was not found. Examples of symbols are global / constant variable names, driver function names, texture names, and surface names."""
cudaErrorNotReady = 600
"""This indicates that asynchronous operations issued previously have not completed yet. This result is not actually an error, but must be indicated differently than :data:`cudaSuccess` (which indicates completion). Calls that may return this value include :func:`cudaEventQuery` and :func:`cudaStreamQuery`."""
cudaErrorIllegalAddress = 700
"""The device encountered a load or store instruction on an invalid memory address. This leaves the process in an inconsistent state and any further CUDA work will return the same error. To continue using CUDA, the process must be terminated and relaunched."""
cudaErrorLaunchOutOfResources = 701
"""This indicates that a launch did not occur because it did not have appropriate resources. Although this error is similar to :data:`cudaErrorInvalidConfiguration`, this error usually indicates that the user has attempted to pass too many arguments to the device kernel, or the kernel launch specifies too many threads for the kernel's register count."""
cudaErrorLaunchTimeout = 702
"""This indicates that the device kernel took too long to execute. This can only occur if timeouts are enabled - see the device property kernelExecTimeoutEnabled for more information. This leaves the process in an inconsistent state and any further CUDA work will return the same error. To continue using CUDA, the process must be terminated and relaunched."""
cudaErrorLaunchIncompatibleTexturing = 703
"""This error indicates a kernel launch that uses an incompatible texturing mode."""
cudaErrorPeerAccessAlreadyEnabled = 704
"""This error indicates that a call to :func:`cudaDeviceEnablePeerAccess` is trying to re-enable peer addressing on from a context which has already had peer addressing enabled."""
cudaErrorPeerAccessNotEnabled = 705
"""This error indicates that :func:`cudaDeviceDisablePeerAccess` is trying to disable peer addressing which has not been enabled yet via :func:`cudaDeviceEnablePeerAccess`."""
cudaErrorSetOnActiveProcess = 708
"""This indicates that the user has called :func:`cudaSetValidDevices`, :func:`cudaSetDeviceFlags`, :func:`cudaD3D9SetDirect3DDevice`, :func:`cudaD3D10SetDirect3DDevice`, :func:`cudaD3D11SetDirect3DDevice`, or :func:`cudaVDPAUSetVDPAUDevice` after initializing the CUDA Runtime by calling non-device management operations (allocating memory and launching kernels are examples of non-device management operations). This error can also be returned if using runtime / driver interoperability and there is an existing :class:`CUcontext` active on the host thread."""
cudaErrorContextIsDestroyed = 709
"""This error indicates that the context current to the calling thread has been destroyed using cuCtxDestroy, or is a primary context which has not yet been initialized."""
cudaErrorAssert = 710
"""An assert triggered in device code during kernel execution. The device cannot be used again. All existing allocations are invalid. To continue using CUDA, the process must be terminated and relaunched."""
cudaErrorTooManyPeers = 711
"""This error indicates that the hardware resources required to enable peer access have been exhausted for one or more of the devices passed to :func:`cudaEnablePeerAccess`."""
cudaErrorHostMemoryAlreadyRegistered = 712
"""This error indicates that the memory range passed to :func:`cudaHostRegister` has already been registered."""
cudaErrorHostMemoryNotRegistered = 713
"""This error indicates that the pointer passed to :func:`cudaHostUnregister` does not correspond to any currently registered memory region."""
cudaErrorHardwareStackError = 714
"""Device encountered an error in the call stack during kernel execution, possibly due to stack corruption or exceeding the stack size limit. This leaves the process in an inconsistent state and any further CUDA work will return the same error. To continue using CUDA, the process must be terminated and relaunched."""
cudaErrorIllegalInstruction = 715
"""The device encountered an illegal instruction during kernel execution This leaves the process in an inconsistent state and any further CUDA work will return the same error. To continue using CUDA, the process must be terminated and relaunched."""
cudaErrorMisalignedAddress = 716
"""The device encountered a load or store instruction on a memory address which is not aligned. This leaves the process in an inconsistent state and any further CUDA work will return the same error. To continue using CUDA, the process must be terminated and relaunched."""
cudaErrorInvalidAddressSpace = 717
"""While executing a kernel, the device encountered an instruction which can only operate on memory locations in certain address spaces (global, shared, or local), but was supplied a memory address not belonging to an allowed address space. This leaves the process in an inconsistent state and any further CUDA work will return the same error. To continue using CUDA, the process must be terminated and relaunched."""
cudaErrorInvalidPc = 718
"""The device encountered an invalid program counter. This leaves the process in an inconsistent state and any further CUDA work will return the same error. To continue using CUDA, the process must be terminated and relaunched."""
cudaErrorLaunchFailure = 719
"""An exception occurred on the device while executing a kernel. Common causes include dereferencing an invalid device pointer and accessing out of bounds shared memory. Less common cases can be system specific - more information about these cases can be found in the system specific user guide. This leaves the process in an inconsistent state and any further CUDA work will return the same error. To continue using CUDA, the process must be terminated and relaunched."""
cudaErrorCooperativeLaunchTooLarge = 720
"""This error indicates that the number of blocks launched per grid for a kernel that was launched via either :func:`cudaLaunchCooperativeKernel` or :func:`cudaLaunchCooperativeKernelMultiDevice` exceeds the maximum number of blocks as allowed by :func:`cudaOccupancyMaxActiveBlocksPerMultiprocessor` or :func:`cudaOccupancyMaxActiveBlocksPerMultiprocessorWithFlags` times the number of multiprocessors as specified by the device attribute :func:`cudaDevAttrMultiProcessorCount`."""
cudaErrorNotPermitted = 800
"""This error indicates the attempted operation is not permitted."""
cudaErrorNotSupported = 801
"""This error indicates the attempted operation is not supported on the current system or device."""
cudaErrorSystemNotReady = 802
"""This error indicates that the system is not yet ready to start any CUDA work. To continue using CUDA, verify the system configuration is in a valid state and all required driver daemons are actively running. More information about this error can be found in the system specific user guide."""
cudaErrorSystemDriverMismatch = 803
"""This error indicates that there is a mismatch between the versions of the display driver and the CUDA driver. Refer to the compatibility documentation for supported versions."""
cudaErrorCompatNotSupportedOnDevice = 804
"""This error indicates that the system was upgraded to run with forward compatibility but the visible hardware detected by CUDA does not support this configuration. Refer to the compatibility documentation for the supported hardware matrix or ensure that only supported hardware is visible during initialization via the ``CUDA_VISIBLE_DEVICES`` environment variable."""
cudaErrorMpsConnectionFailed = 805
"""This error indicates that the MPS client failed to connect to the MPS control daemon or the MPS server."""
cudaErrorMpsRpcFailure = 806
"""This error indicates that the remote procedural call between the MPS server and the MPS client failed."""
cudaErrorMpsServerNotReady = 807
"""This error indicates that the MPS server is not ready to accept new MPS client requests. This error can be returned when the MPS server is in the process of recovering from a fatal failure."""
cudaErrorMpsMaxClientsReached = 808
"""This error indicates that the hardware resources required to create MPS client have been exhausted."""
cudaErrorMpsMaxConnectionsReached = 809
"""This error indicates the the hardware resources required to device connections have been exhausted."""
cudaErrorMpsClientTerminated = 810
"""This error indicates that the MPS client has been terminated by the server. To continue using CUDA, the process must be terminated and relaunched."""
cudaErrorCdpNotSupported = 811
"""This error indicates, that the program is using CUDA Dynamic Parallelism, but the current configuration, like MPS, does not support it."""
cudaErrorCdpVersionMismatch = 812
"""This error indicates, that the program contains an unsupported interaction between different versions of CUDA Dynamic Parallelism."""
cudaErrorStreamCaptureUnsupported = 900
"""The operation is not permitted when the stream is capturing."""
cudaErrorStreamCaptureInvalidated = 901
"""The current capture sequence on the stream has been invalidated due to a previous error."""
cudaErrorStreamCaptureMerge = 902
"""The operation would have resulted in a merge of two independent capture sequences."""
cudaErrorStreamCaptureUnmatched = 903
"""The capture was not initiated in this stream."""
cudaErrorStreamCaptureUnjoined = 904
"""The capture sequence contains a fork that was not joined to the primary stream."""
cudaErrorStreamCaptureIsolation = 905
"""A dependency would have been created which crosses the capture sequence boundary. Only implicit in-stream ordering dependencies are allowed to cross the boundary."""
cudaErrorStreamCaptureImplicit = 906
"""The operation would have resulted in a disallowed implicit dependency on a current capture sequence from :data:`cudaStreamLegacy`."""
cudaErrorCapturedEvent = 907
"""The operation is not permitted on an event which was last recorded in a capturing stream."""
cudaErrorStreamCaptureWrongThread = 908
"""A stream capture sequence not initiated with the :data:`cudaStreamCaptureModeRelaxed` argument to :func:`cudaStreamBeginCapture` was passed to :func:`cudaStreamEndCapture` in a different thread."""
cudaErrorTimeout = 909
"""This indicates that the wait operation has timed out."""
cudaErrorGraphExecUpdateFailure = 910
"""This error indicates that the graph update was not performed because it included changes which violated constraints specific to instantiated graph update."""
cudaErrorExternalDevice = 911
"""This indicates that an async error has occurred in a device outside of CUDA. If CUDA was waiting for an external device's signal before consuming shared data, the external device signaled an error indicating that the data is not valid for consumption. This leaves the process in an inconsistent state and any further CUDA work will return the same error. To continue using CUDA, the process must be terminated and relaunched."""
cudaErrorInvalidClusterSize = 912
"""This indicates that a kernel launch error has occurred due to cluster misconfiguration."""
cudaErrorUnknown = 999
"""This indicates that an unknown internal error has occurred."""
# pylint: enable=line-too-long


# Error Checking #
class cudaError(Exception):
    """Base exception class for CUDA driver query errors."""

    _value_class_mapping: _ClassVar[dict[int, type[cudaError]]] = {}
    _errcode_to_string: _ClassVar[dict[int, str]] = {  # List of currently known error codes
        cudaErrorInitializationError:        'Initialization error.',
        cudaErrorSymbolNotFound:             'Named symbol not found.',
        cudaErrorInvalidValue:               'Invalid argument.',
        cudaErrorNoDevice:                   'No CUDA-capable device is detected.',
        cudaErrorInvalidDevice:              'Invalid device ordinal.',
        cudaErrorSystemDriverMismatch:       'System has unsupported display driver / CUDA driver combination.',
        cudaErrorCudartUnloading:            'Driver shutting down.',
        cudaErrorCompatNotSupportedOnDevice: 'Forward compatibility was attempted on non supported Hardware.',
        cudaErrorDeviceUninitialized:        'Invalid device context.',
    }  # fmt:skip
    _errcode_to_name: _ClassVar[dict[int, str]] = {}
    value: int

    def __new__(cls, value: int) -> _Self:
        """Map value to a proper subclass of :class:`cudaError`."""
        if cls is cudaError:
            # pylint: disable-next=self-cls-assignment
            cls = cudaError._value_class_mapping.get(value, cls)  # type: ignore[assignment]
        obj = Exception.__new__(cls)
        obj.value = value
        return obj

    def __repr__(self) -> str:
        """Return a string representation of the error."""
        # pylint: disable=no-member
        try:
            if self.value not in cudaError._errcode_to_string:
                cudaError._errcode_to_string[self.value] = '{}.'.format(
                    cuGetErrorString(self.value).rstrip('.').capitalize(),
                )
            if self.value not in cudaError._errcode_to_name:
                cudaError._errcode_to_name[self.value] = cudaGetErrorName(self.value)
            return '{} Code: {} ({}).'.format(
                cudaError._errcode_to_string[self.value],
                cudaError._errcode_to_name[self.value],
                self.value,
            )
        except cudaError:
            return f'CUDA Error with code {self.value}.'

    def __eq__(self, other: object) -> bool:
        """Test equality to other object."""
        if not isinstance(other, cudaError):
            return NotImplemented
        return self.value == other.value  # pylint: disable=no-member

    def __reduce__(self) -> tuple[type[cudaError], tuple[int]]:
        """Return state information for pickling."""
        return cudaError, (self.value,)  # pylint: disable=no-member


def cudaExceptionClass(cudaErrorCode: int) -> type[cudaError]:
    """Map value to a proper subclass of :class:`cudaError`.

    Raises:
        ValueError: If the error code is not valid.
    """
    if cudaErrorCode not in cudaError._value_class_mapping:  # pylint: disable=protected-access
        raise ValueError(f'cudaErrorCode {cudaErrorCode} is not valid.')
    return cudaError._value_class_mapping[cudaErrorCode]  # pylint: disable=protected-access


def _extract_cuda_errors_as_classes() -> None:
    """Generate a hierarchy of classes on top of :class:`cudaError` class.

    Each CUDA Error gets a new :class:`cudaError` subclass. This way try-except blocks can filter
    appropriate exceptions more easily.

    :class:`cudaError` is a parent class. Each ``cudaError*`` gets it's own subclass.
    e.g. :data:`cudaErrorInvalidValue` will be turned into :class:`cudaError_InvalidValue`.
    """
    this_module = _sys.modules[__name__]
    cuda_error_names = [
        x
        for x in dir(this_module)
        if x.startswith('cudaError') and not x.startswith('cudaError_') and x != 'cudaError'
    ]
    for err_name in cuda_error_names:
        # e.g. Turn cudaErrorInvalidValue into cudaError_InvalidValue
        class_name = err_name.replace('cudaError', 'cudaError_')
        err_val = getattr(this_module, err_name)

        def gen_new(value: int) -> _Callable[[type[cudaError]], cudaError]:
            def new(cls: type[cudaError]) -> cudaError:
                return cudaError.__new__(cls, value)

            return new

        # pylint: disable=protected-access
        new_error_class = type(class_name, (cudaError,), {'__new__': gen_new(err_val)})
        new_error_class.__module__ = __name__
        if err_val in cudaError._errcode_to_string:
            new_error_class.__doc__ = 'cudaError: {} Code: :data:`{}` ({}).'.format(
                cudaError._errcode_to_string[err_val],
                err_name,
                err_val,
            )
        else:
            new_error_class.__doc__ = f'CUDA Error with code :data:`{err_name}` ({err_val})'
        setattr(this_module, class_name, new_error_class)
        cudaError._value_class_mapping[err_val] = new_error_class
        cudaError._errcode_to_name[err_val] = err_name


# Add explicit references to appease linters
class __cudaError(cudaError):
    value: int

    def __new__(cls) -> cudaError:  # type: ignore[misc,empty-body]
        ...


cudaError_InitializationError: type[__cudaError]
cudaError_SymbolNotFound: type[__cudaError]
cudaError_InvalidValue: type[__cudaError]
cudaError_NoDevice: type[__cudaError]
cudaError_InvalidDevice: type[__cudaError]
cudaError_SystemDriverMismatch: type[__cudaError]
cudaError_CudartUnloading: type[__cudaError]
cudaError_CompatNotSupportedOnDevice: type[__cudaError]
cudaError_DeviceUninitialized: type[__cudaError]

_extract_cuda_errors_as_classes()
del _extract_cuda_errors_as_classes


def _cudaCheckReturn(ret: _Any) -> _Any:
    if ret != cudaSuccess:
        raise cudaError(ret)
    return ret


# Function access #
__cudaLib: _ctypes.CDLL | None = None
__libLoadLock: _threading.Lock = _threading.Lock()
# Function pointers are cached to prevent unnecessary libLoadLock locking
__cudaGetFunctionPointer_cache: dict[str, _ctypes._CFuncPtr] = {}  # type: ignore[name-defined]


def __cudaGetFunctionPointer(name: str) -> _ctypes._CFuncPtr:  # type: ignore[name-defined]
    """Get the function pointer from the CUDA Runtime library.

    Raises:
        cudaError_InitializationError:
            If cannot found the CUDA Runtime library.
        cudaError_SymbolNotFound:
            If cannot found the function pointer.
    """
    if name in __cudaGetFunctionPointer_cache:
        return __cudaGetFunctionPointer_cache[name]

    if __cudaLib is None:
        __LoadCudaLibrary()

    with __libLoadLock:
        try:
            __cudaGetFunctionPointer_cache[name] = getattr(__cudaLib, name)
            return __cudaGetFunctionPointer_cache[name]
        except AttributeError as ex:
            raise cudaError(cudaErrorSymbolNotFound) from ex


def __LoadCudaLibrary() -> None:  # pylint: disable=too-many-branches
    """Load the library if it isn't loaded already.

    Raises:
        cudaError_InitializationError:
            If cannot found the CUDA Runtime library.
    """
    global __cudaLib  # pylint: disable=global-statement

    if __cudaLib is None:
        # Lock to ensure only one caller loads the library
        with __libLoadLock:
            # Ensure the library still isn't loaded
            if __cudaLib is None:  # pylint: disable=too-many-nested-blocks
                # Platform specific libcudart location
                system = _platform.system()
                bits = _platform.architecture()[0].replace('bit', '')
                if system == 'Darwin':
                    lib_filenames = ['libcudart.dylib']
                elif system == 'Linux':
                    lib_filenames = ['libcudart.so']
                elif system == 'Windows':
                    lib_filenames = [f'cudart{bits}.dll', 'cudart.dll']

                # Open library
                for lib_filename in lib_filenames:
                    try:
                        __cudaLib = _ctypes.CDLL(lib_filename)
                        break
                    except OSError:
                        pass

                # Try to load the library from the CUDA_PATH environment variable
                if __cudaLib is None:
                    cuda_paths = [
                        _os.getenv(env_name, '')
                        for env_name in ('CUDA_PATH', 'CUDA_HOME', 'CUDA_ROOT')
                    ]
                    if system != 'Windows':
                        cuda_paths.append('/usr/local/cuda')
                        candidate_paths = []
                        for cuda_path in cuda_paths:
                            if _os.path.isdir(cuda_path):
                                for lib_filename in lib_filenames:
                                    candidate_paths.extend(
                                        [
                                            _os.path.join(cuda_path, f'lib{bits}', lib_filename),
                                            _os.path.join(cuda_path, 'lib', lib_filename),
                                        ],
                                    )
                    else:
                        candidate_dirs = _os.getenv('PATH', '').split(_os.path.pathsep)
                        candidate_paths = []
                        for cuda_path in cuda_paths:
                            if _os.path.isdir(cuda_path):
                                candidate_dirs.extend(
                                    [
                                        _os.path.join(cuda_path, 'bin'),
                                        _os.path.join(cuda_path, f'lib{bits}'),
                                        _os.path.join(cuda_path, 'lib'),
                                    ],
                                )
                        for candidate_dir in candidate_dirs:
                            candidate_paths.extend(
                                _glob.iglob(_os.path.join(candidate_dir, f'cudart{bits}*.dll')),
                            )

                    # Normalize paths and remove duplicates
                    candidate_paths = list(
                        dict.fromkeys(
                            _os.path.normpath(_os.path.normcase(p)) for p in candidate_paths
                        ),
                    )
                    for lib_filename in candidate_paths:
                        try:
                            __cudaLib = _ctypes.CDLL(lib_filename)
                            break
                        except OSError:
                            pass

                if __cudaLib is None:
                    _cudaCheckReturn(cudaErrorInitializationError)


def cudaGetErrorName(error: int) -> str:
    """Get the string representation of an error code enum name.

    Returns: str
        A string containing the name of an error code in the enum. If the error code is not
        recognized, "unrecognized error code" is returned.

    Raises:
        cudaError_InitializationError:
            If cannot found the CUDA Runtime library.
    """
    fn = __cudaGetFunctionPointer('cudaGetErrorName')

    fn.restype = _ctypes.c_char_p  # otherwise return is an int
    p_name = fn(_cudaError_t(error))
    name = _ctypes.string_at(p_name)
    return name.decode('utf-8', errors='replace')


def cuGetErrorString(error: int) -> str:
    """Get the description string for an error code.

    Returns: str
        The description string for an error code. If the error code is not recognized, "unrecognized
        error code" is returned.

    Raises:
        cudaError_InitializationError:
            If cannot found the CUDA Runtime library.
    """
    fn = __cudaGetFunctionPointer('cudaGetErrorString')

    fn.restype = _ctypes.c_char_p  # otherwise return is an int
    p_name = fn(_cudaError_t(error))
    name = _ctypes.string_at(p_name)
    return name.decode('utf-8', errors='replace')


def cudaGetLastError() -> int:
    """Get the last error from a runtime call.

    Returns: int
        The last error that has been produced by any of the runtime calls in the same instance of
        the CUDA Runtime library in the host thread and resets it to :data:`cudaSuccess`.

    Raises:
        cudaError_InitializationError:
            If cannot found the CUDA Runtime library.
        cudaError_InsufficientDriver:
            If the installed NVIDIA CUDA driver is older than the CUDA Runtime library.
        cudaError_NoDevice:
            If no CUDA-capable devices were detected by the installed CUDA driver.
    """
    fn = __cudaGetFunctionPointer('cudaGetLastError')
    return fn()


def cudaPeekAtLastError() -> int:
    """Get the last error from a runtime call.

    Returns: int
        The last error that has been produced by any of the runtime calls in the same instance of
        the CUDA Runtime library in the host thread. This call does not reset the error to
        :data:`cudaSuccess` like :func:`cudaGetLastError`.

    Raises:
        cudaError_InitializationError:
            If cannot found the CUDA Runtime library.
        cudaError_InsufficientDriver:
            If the installed NVIDIA CUDA driver is older than the CUDA Runtime library.
        cudaError_NoDevice:
            If no CUDA-capable devices were detected by the installed CUDA driver.
    """
    fn = __cudaGetFunctionPointer('cudaPeekAtLastError')
    return fn()


def cudaDriverGetVersion() -> str:
    """Get the latest CUDA version supported by driver.

    Returns: str
        The latest version of CUDA supported by the driver of the form :data:`'<major>.<minor>'`.

    Raises:
        cudaError_InitializationError:
            If cannot found the CUDA Runtime library.
        cudaError_InsufficientDriver:
            If the installed NVIDIA CUDA driver is older than the CUDA Runtime library.
        cudaError_NoDevice:
            If no CUDA-capable devices were detected by the installed CUDA driver.
    """
    fn = __cudaGetFunctionPointer('cudaDriverGetVersion')

    driver_version = _ctypes.c_int()
    ret = fn(_ctypes.byref(driver_version))
    _cudaCheckReturn(ret)
    major = driver_version.value // 1000
    minor = (driver_version.value % 1000) // 10
    return f'{major}.{minor}'


def cudaRuntimeGetVersion() -> str:
    """Get the CUDA Runtime version.

    Returns: str
        The version number of the current CUDA Runtime instance of the form :data:`'<major>.<minor>'`.

    Raises:
        cudaError_InitializationError:
            If cannot found the CUDA Runtime library.
        cudaError_InsufficientDriver:
            If the installed NVIDIA CUDA driver is older than the CUDA Runtime library.
        cudaError_NoDevice:
            If no CUDA-capable devices were detected by the installed CUDA driver.
    """
    fn = __cudaGetFunctionPointer('cudaRuntimeGetVersion')

    runtime_version = _ctypes.c_int()
    ret = fn(_ctypes.byref(runtime_version))
    _cudaCheckReturn(ret)
    major = runtime_version.value // 1000
    minor = (runtime_version.value % 1000) // 10
    return f'{major}.{minor}'


def cudaGetDeviceCount() -> int:
    """Get the number of compute-capable devices.

    Returns: int
        The number of devices with compute capability greater or equal to 2.0 that are available for
        execution.

    Raises:
        cudaError_InitializationError:
            If cannot found the CUDA Runtime library.
        cudaError_InsufficientDriver:
            If the installed NVIDIA CUDA driver is older than the CUDA Runtime library.
        cudaError_NoDevice:
            If no CUDA-capable devices were detected by the installed CUDA driver.
    """
    fn = __cudaGetFunctionPointer('cudaGetDeviceCount')

    count = _ctypes.c_int(0)
    ret = fn(_ctypes.byref(count))
    _cudaCheckReturn(ret)
    return count.value


def cudaDeviceGetByPCIBusId(pciBusId: str) -> int:
    """Get a handle to a compute device.

    Args:
        pciBusId (str):
            String in one of the following forms: ``[domain]:[bus]:[device].[function]``,
            ``[domain]:[bus]:[device]``, ``[bus]:[device].[function]`` where ``domain``, ``bus``,
            ``device``, and ``function`` are all hexadecimal values.

    Returns: int
        A device ordinal given a PCI bus ID string.

    Raises:
        cudaError_InitializationError:
            If cannot found the CUDA Runtime library.
        cudaError_InsufficientDriver:
            If the installed NVIDIA CUDA driver is older than the CUDA Runtime library.
        cudaError_NoDevice:
            If no CUDA-capable devices were detected by the installed CUDA driver.
        cudaError_InvalidValue:
            If the value of :data:`pciBusId` is not a valid PCI bus identifier.
        cudaError_InvalidDevice:
            If the device ordinal supplied by the user does not correspond to a valid CUDA device.
    """
    fn = __cudaGetFunctionPointer('cudaDeviceGetByPCIBusId')

    device = _ctypes.c_int()
    ret = fn(_ctypes.byref(device), _ctypes.c_char_p(pciBusId.encode('utf-8')))
    _cudaCheckReturn(ret)
    return device.value


def cudaDeviceGetPCIBusId(device: int) -> str:
    """Get a PCI Bus Id string for the device.

    Returns: str
        An ASCII string identifying the device.

    Raises:
        cudaError_InitializationError:
            If cannot found the CUDA Runtime library.
        cudaError_InsufficientDriver:
            If the installed NVIDIA CUDA driver is older than the CUDA Runtime library.
        cudaError_NoDevice:
            If no CUDA-capable devices were detected by the installed CUDA driver.
        cudaError_InvalidValue:
            If the value of :data:`device` is not a valid device ordinal.
        cudaError_InvalidDevice:
            If the device ordinal supplied by the user does not correspond to a valid CUDA device.
    """
    fn = __cudaGetFunctionPointer('cudaDeviceGetPCIBusId')

    pciBusId = _ctypes.create_string_buffer(256)
    ret = fn(pciBusId, _ctypes.c_int(256), _ctypes.c_int(device))
    _cudaCheckReturn(ret)
    return pciBusId.value.decode('utf-8', errors='replace')


def is_available() -> bool:
    """Test whether there are any CUDA visible devices."""
    try:
        return cudaGetDeviceCount() > 0
    except cudaError:
        return False



# This file is part of nvitop, the interactive NVIDIA-GPU process viewer.
# License: GNU GPL version 3.

# pylint: disable=missing-module-docstring,missing-class-docstring,missing-function-docstring

import threading
import time

from cachetools.func import ttl_cache

from nvitop.gui.library import NA, Device, Displayable, colored, cut_string, host, make_bar
from nvitop.version import __version__


class DevicePanel(Displayable):  # pylint: disable=too-many-instance-attributes
    NAME = 'device'
    SNAPSHOT_INTERVAL = 0.5

    def __init__(self, devices, compact, win, root):
        super().__init__(win, root)

        self.devices = devices
        self.device_count = len(self.devices)

        self.all_devices = []
        self.leaf_devices = []
        self.mig_device_counts = [0] * self.device_count
        self.mig_enabled_device_count = 0
        for i, device in enumerate(self.devices):
            self.all_devices.append(device)

            mig_devices = device.mig_devices()
            self.mig_device_counts[i] = len(mig_devices)
            if self.mig_device_counts[i] > 0:
                self.all_devices.extend(mig_devices)
                self.leaf_devices.extend(mig_devices)
                self.mig_enabled_device_count += 1
            else:
                self.leaf_devices.append(device)

        self.mig_device_count = sum(self.mig_device_counts)
        self.all_device_count = len(self.all_devices)
        self.leaf_device_count = len(self.leaf_devices)

        self._compact = compact
        self.width = max(79, root.width)
        self.compact_height = (
            4 + 2 * (self.device_count + 1) + self.mig_device_count + self.mig_enabled_device_count
        )
        self.full_height = self.compact_height + self.device_count + 1
        self.height = self.compact_height if compact else self.full_height
        if self.device_count == 0:
            self.height = self.full_height = self.compact_height = 6

        self.driver_version = Device.driver_version()
        self.cuda_driver_version = Device.cuda_driver_version()

        self._snapshot_buffer = []
        self._snapshots = []
        self.snapshot_lock = threading.Lock()
        self.snapshots = self.take_snapshots()
        self._snapshot_daemon = threading.Thread(
            name='device-snapshot-daemon',
            target=self._snapshot_target,
            daemon=True,
        )
        self._daemon_running = threading.Event()

        self.formats_compact = [
            '│ {physical_index:>3} {fan_speed_string:>3} {temperature_string:>4} '
            '{performance_state:>3} {power_status:>12} '
            '│ {memory_usage:>20} │ {gpu_utilization_string:>7}  {compute_mode:>11} │',
        ]
        self.formats_full = [
            '│ {physical_index:>3}  {name:<18}  {persistence_mode:<4} '
            '│ {bus_id:<16} {display_active:>3} │ {total_volatile_uncorrected_ecc_errors:>20} │',
            '│ {fan_speed_string:>3}  {temperature_string:>4}  {performance_state:>4}  {power_status:>12} '
            '│ {memory_usage:>20} │ {gpu_utilization_string:>7}  {compute_mode:>11} │',
        ]

        self.mig_formats = [
            '│{physical_index:>2}:{mig_index:<2}{name:>12} @ GI/CI:{gpu_instance_id:>2}/{compute_instance_id:<2}'
            '│ {memory_usage:>20} │ BAR1: {bar1_memory_used_human:>8} / {bar1_memory_percent_string:>3} │',
        ]

        if host.WINDOWS:
            self.formats_full[0] = self.formats_full[0].replace(
                'persistence_mode',
                'current_driver_model',
            )

        self.support_mig = any('N/A' not in device.mig_mode for device in self.snapshots)
        if self.support_mig:
            self.formats_full[0] = self.formats_full[0].replace(
                '{total_volatile_uncorrected_ecc_errors:>20}',
                '{mig_mode:>8}  {total_volatile_uncorrected_ecc_errors:>10}',
            )

    @property
    def width(self):
        return self._width

    @width.setter
    def width(self, value):
        width = max(79, value)
        if self._width != width and self.visible:
            self.need_redraw = True
        self._width = width

    @property
    def compact(self):
        return self._compact

    @compact.setter
    def compact(self, value):
        if self._compact != value:
            self.need_redraw = True
            self._compact = value
            self.height = self.compact_height if self.compact else self.full_height

    @property
    def snapshots(self):
        return self._snapshots

    @snapshots.setter
    def snapshots(self, snapshots):
        with self.snapshot_lock:
            self._snapshots = snapshots

    @classmethod
    def set_snapshot_interval(cls, interval):
        assert interval > 0.0
        interval = float(interval)

        cls.SNAPSHOT_INTERVAL = min(interval / 3.0, 1.0)
        cls.take_snapshots = ttl_cache(ttl=interval)(
            cls.take_snapshots.__wrapped__,  # pylint: disable=no-member
        )

    @ttl_cache(ttl=1.0)
    def take_snapshots(self):
        snapshots = [device.as_snapshot() for device in self.all_devices]

        for device in snapshots:
            if device.name.startswith('NVIDIA '):
                device.name = device.name.replace('NVIDIA ', '', 1)
            if device.is_mig_device:
                device.name = device.name.rpartition(' ')[-1]
                if device.bar1_memory_percent is not NA:
                    device.bar1_memory_percent = round(device.bar1_memory_percent)
                    if device.bar1_memory_percent >= 100:
                        device.bar1_memory_percent_string = 'MAX'
                    else:
                        device.bar1_memory_percent_string = f'{round(device.bar1_memory_percent)}%'
            else:
                device.name = cut_string(device.name, maxlen=18, padstr='..', align='right')
                device.current_driver_model = device.current_driver_model.replace('WDM', 'TCC')
                device.display_active = device.display_active.replace('Enabled', 'On').replace(
                    'Disabled',
                    'Off',
                )
                device.persistence_mode = device.persistence_mode.replace('Enabled', 'On').replace(
                    'Disabled',
                    'Off',
                )
                device.mig_mode = device.mig_mode.replace('N/A', 'N/A ')
                device.compute_mode = device.compute_mode.replace('Exclusive', 'E.')
                if device.fan_speed is not NA and device.fan_speed >= 100:
                    device.fan_speed_string = 'MAX'

        with self.snapshot_lock:
            self._snapshot_buffer = snapshots

        return snapshots

    def _snapshot_target(self):
        self._daemon_running.wait()
        while self._daemon_running.is_set():
            self.take_snapshots()
            time.sleep(self.SNAPSHOT_INTERVAL)

    def header_lines(self, compact=None):
        if compact is None:
            compact = self.compact

        version_infos = [
            'NVITOP {}'.format(__version__.partition('+')[0]),
            f'Driver Version: {self.driver_version}',
            f'CUDA Driver Version: {self.cuda_driver_version}',
        ]
        if sum(len(v) for v in version_infos) % 2 == 0:
            version_infos[0] += ' '
        version_seps = ' ' * max(2, (75 - sum(len(v) for v in version_infos)) // 2)

        header = [
            '╒═════════════════════════════════════════════════════════════════════════════╕',
            '│ {} │'.format(version_seps.join(version_infos).ljust(75, ' ')),
        ]
        if self.device_count > 0:
            header.append(
                '├───────────────────────────────┬──────────────────────┬──────────────────────┤',
            )
            if compact:
                header.append(
                    '│ GPU Fan Temp Perf Pwr:Usg/Cap │         Memory-Usage │ GPU-Util  Compute M. │',
                )
            else:
                header.extend(
                    (
                        '│ GPU  Name        Persistence-M│ Bus-Id        Disp.A │ Volatile Uncorr. ECC │',
                        '│ Fan  Temp  Perf  Pwr:Usage/Cap│         Memory-Usage │ GPU-Util  Compute M. │',
                    ),
                )
                if host.WINDOWS:
                    header[-2] = header[-2].replace('Persistence-M', '    TCC/WDDM ')
                if self.support_mig:
                    header[-2] = header[-2].replace('Volatile Uncorr. ECC', 'MIG M.   Uncorr. ECC')
            header.append(
                '╞═══════════════════════════════╪══════════════════════╪══════════════════════╡',
            )
        else:
            header.extend(
                (
                    '╞═════════════════════════════════════════════════════════════════════════════╡',
                    '│  No visible devices found                                                   │',
                    '╘═════════════════════════════════════════════════════════════════════════════╛',
                ),
            )
        return header

    def frame_lines(self, compact=None):
        if compact is None:
            compact = self.compact

        frame = self.header_lines(compact)

        remaining_width = self.width - 79
        data_line = (
            '│                               │                      │                      │'
        )
        separator_line = (
            '├───────────────────────────────┼──────────────────────┼──────────────────────┤'
        )
        if self.width >= 100:
            data_line += ' ' * (remaining_width - 1) + '│'
            separator_line = separator_line[:-1] + '┼' + '─' * (remaining_width - 1) + '┤'

        if self.device_count > 0:
            for mig_device_count in self.mig_device_counts:
                if compact:
                    frame.extend((data_line, separator_line))
                else:
                    frame.extend((data_line, data_line, separator_line))
                if mig_device_count > 0:
                    frame.extend((data_line,) * mig_device_count + (separator_line,))

            frame.pop()
            frame.append(
                '╘═══════════════════════════════╧══════════════════════╧══════════════════════╛',
            )
            if self.width >= 100:
                frame[5 - int(compact)] = (
                    frame[5 - int(compact)][:-1] + '╪' + '═' * (remaining_width - 1) + '╕'
                )
                frame[-1] = frame[-1][:-1] + '╧' + '═' * (remaining_width - 1) + '╛'

        return frame

    def poke(self):
        if not self._daemon_running.is_set():
            self._daemon_running.set()
            self._snapshot_daemon.start()

        self.snapshots = self._snapshot_buffer

        super().poke()

    # pylint: disable-next=too-many-locals,too-many-branches,too-many-statements
    def draw(self):
        self.color_reset()

        if self.need_redraw:
            self.addstr(self.y, self.x, '(Press h for help or q to quit)'.rjust(79))
            self.color_at(self.y, self.x + 55, width=1, fg='magenta', attr='bold | italic')
            self.color_at(self.y, self.x + 69, width=1, fg='magenta', attr='bold | italic')
            for y, line in enumerate(self.frame_lines(), start=self.y + 1):
                self.addstr(y, self.x, line)

        self.addstr(self.y, self.x, cut_string(time.strftime('%a %b %d %H:%M:%S %Y'), maxlen=32))

        formats = self.formats_compact if self.compact else self.formats_full

        remaining_width = self.width - 79
        draw_bars = self.width >= 100
        try:
            selected_device = self.parent.selection.process.device
        except AttributeError:
            selected_device = None

        y_start = self.y + len(formats) + 5
        prev_device_index = self.snapshots[0].tuple_index
        for index, device in enumerate(self.snapshots):  # pylint: disable=too-many-nested-blocks
            if (
                len(prev_device_index) != len(device.tuple_index)
                or prev_device_index[0] != device.tuple_index[0]
            ):
                y_start += 1

            attr = 0
            if selected_device is not None:
                if device.real == selected_device:
                    attr = 'bold'
                elif (
                    device.is_mig_device or device.physical_index != selected_device.physical_index
                ):
                    attr = 'dim'

            fmts = self.mig_formats if device.is_mig_device else formats
            for y, fmt in enumerate(fmts, start=y_start):
                self.addstr(y, self.x, fmt.format(**device.__dict__))
                self.color_at(y, self.x + 1, width=31, fg=device.display_color, attr=attr)
                self.color_at(y, self.x + 33, width=22, fg=device.display_color, attr=attr)
                self.color_at(y, self.x + 56, width=22, fg=device.display_color, attr=attr)

            if draw_bars:
                matrix = [
                    (
                        self.x + 80,
                        y_start,
                        remaining_width - 3,
                        'MEM',
                        device.memory_percent,
                        device.memory_display_color,
                    ),
                    (
                        self.x + 80,
                        y_start + 1,
                        remaining_width - 3,
                        'UTL',
                        device.gpu_utilization,
                        device.gpu_display_color,
                    ),
                ]
                if self.compact:
                    if remaining_width >= 44 and not device.is_mig_device:
                        left_width = (remaining_width - 6 + 1) // 2 - 1
                        right_width = (remaining_width - 6) // 2 + 1
                        matrix = [
                            (
                                self.x + 80,
                                y_start,
                                left_width,
                                'MEM',
                                device.memory_percent,
                                device.memory_display_color,
                            ),
                            (
                                self.x + 80 + left_width + 3,
                                y_start,
                                right_width,
                                'UTL',
                                device.gpu_utilization,
                                device.gpu_display_color,
                            ),
                        ]
                        separator = '┼' if index > 0 else '╤'
                        if len(prev_device_index) == 2:
                            separator = '┬'
                        self.addstr(y_start - 1, self.x + 80 + left_width + 1, separator)
                        self.addstr(y_start, self.x + 80 + left_width + 1, '│')
                        if index == len(self.snapshots) - 1:
                            self.addstr(y_start + 1, self.x + 80 + left_width + 1, '╧')
                    else:
                        if remaining_width >= 44 and len(prev_device_index) == 1:
                            self.addstr(y_start - 1, self.x + 80 + left_width + 1, '┴')
                        matrix.pop()
                elif device.is_mig_device:
                    matrix.pop()
                for x_offset, y, width, prefix, utilization, color in matrix:
                    # pylint: disable-next=disallowed-name
                    bar = make_bar(prefix, utilization, width=width)
                    self.addstr(y, x_offset, bar)
                    if self.TERM_256COLOR:
                        parts = bar.rstrip().split(' ')
                        prefix_len = len(parts[0])
                        bar_len = len(parts[1])
                        full_bar_len = width - prefix_len - 5
                        self.color_at(y, x_offset, width=width, fg=float(bar_len / full_bar_len))
                        for i, x in enumerate(
                            range(x_offset + prefix_len + 1, x_offset + prefix_len + 1 + bar_len),
                        ):
                            self.color_at(y, x, width=1, fg=float(i / full_bar_len))
                    else:
                        self.color_at(y, x_offset, width=width, fg=color, attr=attr)

            y_start += len(fmts)
            prev_device_index = device.tuple_index

    def destroy(self):
        super().destroy()
        self._daemon_running.clear()

    def print_width(self):
        if self.device_count > 0 and self.width >= 100:
            return self.width
        return 79

    def print(self):  # pylint: disable=too-many-locals,too-many-branches
        lines = [time.strftime('%a %b %d %H:%M:%S %Y'), *self.header_lines(compact=False)]

        if self.device_count > 0:
            prev_device_index = self.snapshots[0].tuple_index
            for device in self.snapshots:
                if (
                    len(prev_device_index) != len(device.tuple_index)
                    or prev_device_index[0] != device.tuple_index[0]
                ):
                    lines.append(
                        '├───────────────────────────────┼──────────────────────┼──────────────────────┤',
                    )

                def colorize(s):
                    if len(s) > 0:
                        # pylint: disable-next=cell-var-from-loop
                        return colored(s, device.display_color)  # noqa: B023
                    return ''

                fmts = self.mig_formats if device.is_mig_device else self.formats_full
                for fmt in fmts:
                    line = fmt.format(**device.__dict__)
                    lines.append('│'.join(map(colorize, line.split('│'))))

                prev_device_index = device.tuple_index

            lines.append(
                '╘═══════════════════════════════╧══════════════════════╧══════════════════════╛',
            )

            if self.width >= 100:
                remaining_width = self.width - 79
                y_start = 7
                prev_device_index = self.snapshots[0].tuple_index
                for index, device in enumerate(self.snapshots):
                    if (
                        len(prev_device_index) != len(device.tuple_index)
                        or prev_device_index[0] != device.tuple_index[0]
                    ):
                        lines[y_start] = (
                            lines[y_start][:-1] + '┼' + '─' * (remaining_width - 1) + '┤'
                        )
                        y_start += 1
                    elif index == 0:
                        lines[y_start - 1] = (
                            lines[y_start - 1][:-1] + '╪' + '═' * (remaining_width - 1) + '╕'
                        )

                    matrix = [
                        (
                            'MEM',
                            device.memory_percent,
                            device.memory_display_color,
                        ),
                        (
                            'UTL',
                            device.gpu_utilization,
                            device.gpu_display_color,
                        ),
                    ]
                    if device.is_mig_device:
                        matrix.pop()
                    for y, (prefix, utilization, color) in enumerate(matrix, start=y_start):
                        bar = make_bar(  # pylint: disable=disallowed-name
                            prefix,
                            utilization,
                            width=remaining_width - 3,
                        )
                        lines[y] += f' {colored(bar, color)} │'

                    if index == len(self.snapshots) - 1:
                        lines[y_start + len(matrix)] = (
                            lines[y_start + len(matrix)][:-1]
                            + '╧'
                            + '═' * (remaining_width - 1)
                            + '╛'
                        )

                    y_start += len(matrix)
                    prev_device_index = device.tuple_index

        lines = '\n'.join(lines)
        if self.ascii:
            lines = lines.translate(self.ASCII_TRANSTABLE)

        try:
            print(lines)
        except UnicodeError:
            print(lines.translate(self.ASCII_TRANSTABLE))

    def press(self, key):
        self.root.keymaps.use_keymap('device')
        self.root.press(key)



# This file is part of nvitop, the interactive NVIDIA-GPU process viewer.
# License: GNU GPL version 3.

# pylint: disable=missing-module-docstring,missing-class-docstring,missing-function-docstring

import threading
from functools import partial

from nvitop.gui.library import LARGE_INTEGER, DisplayableContainer, MouseEvent, send_signal
from nvitop.gui.screens.main.device import DevicePanel
from nvitop.gui.screens.main.host import HostPanel
from nvitop.gui.screens.main.process import ProcessPanel


class BreakLoop(Exception):  # noqa: N818
    pass


class MainScreen(DisplayableContainer):  # pylint: disable=too-many-instance-attributes
    NAME = 'main'

    # pylint: disable-next=redefined-builtin,too-many-arguments,too-many-locals,too-many-statements
    def __init__(self, devices, filters, ascii, mode, win, root):
        super().__init__(win, root)

        self.width = root.width

        assert mode in ('auto', 'full', 'compact')
        compact = mode == 'compact'
        self.mode = mode
        self._compact = compact

        self.devices = devices
        self.device_count = len(self.devices)

        self.snapshot_lock = threading.Lock()

        self.device_panel = DevicePanel(self.devices, compact, win=win, root=root)
        self.device_panel.focused = False
        self.add_child(self.device_panel)

        self.host_panel = HostPanel(self.device_panel.leaf_devices, compact, win=win, root=root)
        self.host_panel.focused = False
        self.add_child(self.host_panel)

        self.process_panel = ProcessPanel(
            self.device_panel.leaf_devices,
            compact,
            filters,
            win=win,
            root=root,
        )
        self.process_panel.focused = False
        self.add_child(self.process_panel)

        self.selection = self.process_panel.selection

        self.ascii = ascii
        self.device_panel.ascii = self.ascii
        self.host_panel.ascii = self.ascii
        self.process_panel.ascii = self.ascii
        if ascii:
            self.host_panel.full_height = self.host_panel.height = self.host_panel.compact_height

        self.x, self.y = root.x, root.y
        self.device_panel.x = self.host_panel.x = self.process_panel.x = self.x
        self.device_panel.y = self.y
        self.host_panel.y = self.device_panel.y + self.device_panel.height
        self.process_panel.y = self.host_panel.y + self.host_panel.height
        self.height = self.device_panel.height + self.host_panel.height + self.process_panel.height

    @property
    def compact(self):
        return self._compact

    @compact.setter
    def compact(self, value):
        if self._compact != value:
            self.need_redraw = True
            self._compact = value

    def update_size(self, termsize=None):
        n_term_lines, n_term_cols = termsize = super().update_size(termsize=termsize)

        self.width = n_term_cols - self.x
        self.device_panel.width = self.width
        self.host_panel.width = self.width
        self.process_panel.width = self.width

        self.y = min(self.y, self.root.y)
        height = n_term_lines - self.y
        heights = [
            self.device_panel.full_height
            + self.host_panel.full_height
            + self.process_panel.full_height,
            self.device_panel.compact_height
            + self.host_panel.full_height
            + self.process_panel.full_height,
            self.device_panel.compact_height
            + self.host_panel.compact_height
            + self.process_panel.full_height,
        ]
        if self.mode == 'auto':
            self.compact = height < heights[0]
            self.host_panel.compact = height < heights[1]
            self.process_panel.compact = height < heights[-1]
        else:
            self.compact = self.mode == 'compact'
            self.host_panel.compact = self.compact
            self.process_panel.compact = self.compact
        self.device_panel.compact = self.compact

        self.device_panel.y = self.y
        self.host_panel.y = self.device_panel.y + self.device_panel.height
        self.process_panel.y = self.host_panel.y + self.host_panel.height
        height = self.device_panel.height + self.host_panel.height + self.process_panel.height

        if self.y < self.root.y and self.y + height < n_term_lines:
            self.y = min(self.root.y + self.root.height - height, self.root.y)
            self.update_size(termsize)
            self.need_redraw = True

        if self.height != height:
            self.height = height
            self.need_redraw = True

        return termsize

    def move(self, direction=0):
        if direction == 0:
            return

        self.y -= direction
        self.update_size()
        self.need_redraw = True

    def poke(self):
        super().poke()

        height = self.device_panel.height + self.host_panel.height + self.process_panel.height
        if self.height != height:
            self.update_size()
            self.need_redraw = True

    def draw(self):
        self.color_reset()

        super().draw()

    def print(self):
        if self.device_count > 0:
            print_width = min(panel.print_width() for panel in self.container)
            self.width = max(print_width, min(self.width, 100))
        else:
            self.width = 79
        for panel in self.container:
            panel.width = self.width
            panel.print()

    def __contains__(self, item):
        if self.visible and isinstance(item, MouseEvent):
            return True
        return super().__contains__(item)

    def init_keybindings(self):
        # pylint: disable=too-many-locals,too-many-statements

        def quit():  # pylint: disable=redefined-builtin
            raise BreakLoop

        def change_mode(mode):
            self.mode = mode
            self.root.update_size()

        def force_refresh():
            select_clear()
            host_begin()
            self.y = self.root.y
            self.root.update_size()
            self.root.need_redraw = True

        def screen_move(direction):
            self.move(direction)

        def host_left():
            self.process_panel.host_offset -= 2

        def host_right():
            self.process_panel.host_offset += 2

        def host_begin():
            self.process_panel.host_offset = -1

        def host_end():
            self.process_panel.host_offset = LARGE_INTEGER

        def select_move(direction):
            self.selection.move(direction=direction)

        def select_clear():
            self.selection.clear()

        def tag():
            self.selection.tag()
            select_move(direction=+1)

        def sort_by(order, reverse):
            self.process_panel.order = order
            self.process_panel.reverse = reverse
            self.root.update_size()

        def order_previous():
            sort_by(order=ProcessPanel.ORDERS[self.process_panel.order].previous, reverse=False)

        def order_next():
            sort_by(order=ProcessPanel.ORDERS[self.process_panel.order].next, reverse=False)

        def order_reverse():
            sort_by(order=self.process_panel.order, reverse=not self.process_panel.reverse)

        keymaps = self.root.keymaps

        keymaps.bind('main', 'q', quit)
        keymaps.copy('main', 'q', 'Q')
        keymaps.bind('main', 'a', partial(change_mode, mode='auto'))
        keymaps.bind('main', 'f', partial(change_mode, mode='full'))
        keymaps.bind('main', 'c', partial(change_mode, mode='compact'))
        keymaps.bind('main', 'r', force_refresh)
        keymaps.copy('main', 'r', 'R')
        keymaps.copy('main', 'r', '<C-r>')
        keymaps.copy('main', 'r', '<F5>')

        keymaps.bind('main', '<PageUp>', partial(screen_move, direction=-1))
        keymaps.copy('main', '<PageUp>', '[')
        keymaps.copy('main', '<PageUp>', '<A-K>')
        keymaps.bind('main', '<PageDown>', partial(screen_move, direction=+1))
        keymaps.copy('main', '<PageDown>', ']')
        keymaps.copy('main', '<PageDown>', '<A-J>')

        keymaps.bind('main', '<Left>', host_left)
        keymaps.copy('main', '<Left>', '<A-h>')
        keymaps.bind('main', '<Right>', host_right)
        keymaps.copy('main', '<Right>', '<A-l>')
        keymaps.bind('main', '<C-a>', host_begin)
        keymaps.copy('main', '<C-a>', '^')
        keymaps.bind('main', '<C-e>', host_end)
        keymaps.copy('main', '<C-e>', '$')
        keymaps.bind('main', '<Up>', partial(select_move, direction=-1))
        keymaps.copy('main', '<Up>', '<S-Tab>')
        keymaps.copy('main', '<Up>', '<A-k>')
        keymaps.bind('main', '<Down>', partial(select_move, direction=+1))
        keymaps.copy('main', '<Down>', '<Tab>')
        keymaps.copy('main', '<Down>', '<A-j>')
        keymaps.bind('main', '<Home>', partial(select_move, direction=-(1 << 20)))
        keymaps.bind('main', '<End>', partial(select_move, direction=+(1 << 20)))
        keymaps.bind('main', '<Esc>', select_clear)
        keymaps.bind('main', '<Space>', tag)

        keymaps.bind('main', 'T', partial(send_signal, signal='terminate', panel=self))
        keymaps.bind('main', 'K', partial(send_signal, signal='kill', panel=self))
        keymaps.copy('main', 'K', 'k')
        keymaps.bind('main', '<C-c>', partial(send_signal, signal='interrupt', panel=self))
        keymaps.copy('main', '<C-c>', 'I')

        keymaps.bind('main', ',', order_previous)
        keymaps.copy('main', ',', '<')
        keymaps.bind('main', '.', order_next)
        keymaps.copy('main', '.', '>')
        keymaps.bind('main', '/', order_reverse)
        for order in ProcessPanel.ORDERS:
            keymaps.bind(
                'main',
                'o' + order[:1].lower(),
                partial(sort_by, order=order, reverse=False),
            )
            keymaps.bind(
                'main',
                'o' + order[:1].upper(),
                partial(sort_by, order=order, reverse=True),
            )
